"""This test module exemplifies the basic loop of creating
objective functions without relying on create.

Not using objective_factory.create comes with the added
benefit of having IDE IntelliSense support.
"""


def test_basic_loop_without_create():
    from poli.objective_repository import ToyContinuousProblemFactory

    problem_factory = ToyContinuousProblemFactory()

    problem = problem_factory.create(function_name="ackley_function_01")  # noqa F841


def test_instancing_black_boxes_alone():
    from poli.objective_repository import ToyContinuousBlackBox

    f = ToyContinuousBlackBox(function_name="ackley_function_01")  # noqa F841



"""
This module tests the foldx interface, which can be found in
poli.core.util.proteins.foldx

Since this code is only intended to run inside the
poli__protein environment, we add a skip to the
whole module if the import fails.
"""

from pathlib import Path

import pytest

try:
    from poli.core.util.proteins.foldx import FoldxInterface
except (ImportError, FileNotFoundError):
    pytest.skip("Could not import the foldx interface. ", allow_module_level=True)

try:
    from poli.core.util.proteins.pdb_parsing import (  # noqa F401
        parse_pdb_as_residue_strings,
    )
except ImportError:
    pytest.skip(
        "Could not import the protein utilities for parsing. ", allow_module_level=True
    )

THIS_DIR = Path(__file__).parent.resolve()


@pytest.mark.poli__protein
class TestFoldxInterface:
    wildtype_pdb_path = THIS_DIR / "3ned.pdb"
    tmp_path = THIS_DIR / "tmp"

    foldx_interface = FoldxInterface(tmp_path)

    def test_copying_files(self):
        """
        Tests that the files are copied correctly.
        """
        self.tmp_path.mkdir(exist_ok=True)

        self.foldx_interface.copy_foldx_files(self.wildtype_pdb_path)

        assert (self.foldx_interface.working_dir / "rotabase.txt").exists()
        assert (
            self.foldx_interface.working_dir / f"{self.wildtype_pdb_path.stem}.pdb"
        ).exists()

        # Clean up
        (self.foldx_interface.working_dir / "rotabase.txt").unlink()
        (
            self.foldx_interface.working_dir / f"{self.wildtype_pdb_path.stem}.pdb"
        ).unlink()

    def test_specifying_string_working_dir(self):
        new_foldx_interface = FoldxInterface(str(self.tmp_path))

        new_foldx_interface.copy_foldx_files(self.wildtype_pdb_path)

        assert (new_foldx_interface.working_dir / "rotabase.txt").exists()
        assert (
            new_foldx_interface.working_dir / f"{self.wildtype_pdb_path.stem}.pdb"
        ).exists()

        # Clean up
        (new_foldx_interface.working_dir / "rotabase.txt").unlink()
        (
            new_foldx_interface.working_dir / f"{self.wildtype_pdb_path.stem}.pdb"
        ).unlink()



"""This module implements a batched input generator."""

from itertools import islice
from typing import Iterable


def batched(iterable: Iterable, chunk_size: int):
    """
    Generate batches of elements from an iterable.

    If working on Python 3.12+, this function can be replaced
    with the built-in function `batched` from `itertools`.
    https://docs.python.org/3/library/itertools.html#itertools.batched

    Parameters
    ----------
    iterable : Iterable
        The iterable to be batched.
    chunk_size : int
        The size of each batch.

    Yields
    ------
    tuple
        A tuple containing elements from the iterable, with each tuple having a length equal to chunk_size.

    Examples
    --------
    >>> numbers = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    >>> for batch in batched(numbers, 3):
    ...     print(batch)
    (1, 2, 3)
    (4, 5, 6)
    (7, 8, 9)
    (10,)

    """
    iterator = iter(iterable)
    while chunk := tuple(islice(iterator, chunk_size)):
        yield chunk



import pytest

from poli.core.chemistry.data_packages.random_molecules_data_package import (
    RandomMoleculesDataPackage,
)


@pytest.mark.parametrize("seed", [1, 2, 3])
def test_seed_determinism_for_random_molecule_data_packages(seed):
    data_package_1 = RandomMoleculesDataPackage(
        string_representation="SMILES", seed=seed
    )

    data_package_2 = RandomMoleculesDataPackage(
        string_representation="SMILES", seed=seed
    )

    assert (data_package_1.unsupervised_data == data_package_2.unsupervised_data).all()



from pathlib import Path

import mlflow
import numpy as np

from poli.core.black_box_information import BlackBoxInformation
from poli.core.util.abstract_observer import AbstractObserver

TRACKING_URI = "tracking_uri"
OBJECTIVE = "OBJECTIVE"
SEQUENCE = "SEQUENCE"
SEED = "SEED"


class MLFlowObserver(AbstractObserver):
    """
    This observer uses mlflow as a backend.
    """

    def __init__(self, tracking_uri: Path = None):
        self.step = 0
        if tracking_uri is not None:
            mlflow.set_tracking_uri(tracking_uri)

    def observe(self, x: np.ndarray, y: np.ndarray, context=None) -> None:
        for n in range(y.shape[0]):
            self.step += 1
            mlflow.log_metrics(
                {OBJECTIVE + str(i): y[n, i].item() for i in range(y.shape[1])},
                step=self.step,
            )
            # with mlflow it's unfortunately not so easy to log sequences
            mlflow.log_param(str(self.step) + SEQUENCE, x[n, ...])

    def log(self, algorithm_info: dict):
        mlflow.log_metrics(algorithm_info, step=self.step)

    def initialize_observer(
        self,
        problem_setup_info: BlackBoxInformation,
        caller_info: dict,
        seed: int,
    ) -> object:
        tracking_uri = caller_info.pop(TRACKING_URI, None)
        if tracking_uri is not None:
            mlflow.set_tracking_uri(tracking_uri)

        experiment = mlflow.set_experiment(
            experiment_name=problem_setup_info.get_problem_name()
        )
        run = mlflow.start_run(experiment_id=experiment.experiment_id)
        mlflow.set_tag(SEED, str(seed))
        mlflow.set_tags(caller_info)
        return run.info.run_id

    def finish(self) -> None:
        mlflow.end_run()



import argparse
import os
import subprocess
import tempfile
import time
from pathlib import Path

import Bio.PDB
import Bio.PDB.Polypeptide
import Bio.SeqIO
import openmm.app
import pdbfixer

PDBIO = Bio.PDB.PDBIO()
PDB_PARSER = Bio.PDB.PDBParser(PERMISSIVE=0)


class NonHetSelector(Bio.PDB.Select):
    """Remove HET atoms and choose first conformation of disordered atoms"""

    def accept_residue(self, residue):
        norm_res_bool = residue.get_resname() in [
            pdbfixer.pdbfixer.substitutions[key]
            for key in pdbfixer.pdbfixer.substitutions
        ]
        abnorm_res_bool = residue.get_resname() in [
            key for key in pdbfixer.pdbfixer.substitutions
        ]
        return (
            norm_res_bool or abnorm_res_bool
        )  # Accept abnorm since they are converted later

    def accept_atom(self, atom):
        return (
            not atom.is_disordered()
            or atom.get_altloc() == "A"
            or atom.get_altloc() == "1"
        ) and atom.id[0] in ["C", "H", "N", "O", "S", "P"]


class PDBFixerResIdentifiabilityIssue(Exception):
    pass


def _step_1_reduce(
    reduce_executable,
    pdb_input_filename,
    pdbid,
    temp1,
):
    # Add hydrogens using reduce program
    reduce_executable_path = Path(reduce_executable)
    path_to_wwPDB_het_dict = (
        reduce_executable_path.parent.parent / "reduce_wwPDB_het_dict.txt"
    )
    command = [
        reduce_executable,
        "-BUILD",
        "-DB",
        str(path_to_wwPDB_het_dict),
        "-Quiet",
        pdb_input_filename,
    ]
    error_code = subprocess.Popen(command, stdout=temp1).wait()
    temp1.flush()
    first_model = PDB_PARSER.get_structure(pdbid, temp1.name)[0]

    return first_model


def _step_3_pdbfixer(first_model, temp3):
    for chain in first_model:
        for res in chain:
            for atom in res:
                atom.set_altloc(" ")
    PDBIO.set_structure(first_model)
    PDBIO.save(temp3)
    temp3.flush()

    # Use PDBFixer to fix common PDB errors
    fixer = pdbfixer.PDBFixer(temp3.name)
    fixer.findMissingResidues()
    fixer.findNonstandardResidues()
    fixer.replaceNonstandardResidues()
    fixer.findMissingAtoms()
    fixer.addMissingAtoms()
    fixer.addMissingHydrogens(7.0)
    return temp3, fixer


def _step_4_fix_numbering(fixer, temp3, temp4):
    openmm.app.PDBFile.writeFile(fixer.topology, fixer.positions, temp4, keepIds=False)
    temp4.flush()
    # Fix IDs manually since pdbfixer does not preserve insertion codes
    structure_before = PDB_PARSER.get_structure(temp3.name, temp3.name)
    structure_after = PDB_PARSER.get_structure(temp4.name, temp4.name)
    residues_before = []
    for chain in structure_before[0]:
        residues_before.append(chain.get_list())
    residues_after = []
    for chain in structure_after[0]:
        residues_after.append(chain.get_list())
    chain_counter = ""
    for i, chain in enumerate(structure_before[0]):
        try:
            if (
                structure_after[0].get_list()[i].id
                != structure_before[0].get_list()[i].id
            ):
                try:
                    # HACK BECAUSE OF https://github.com/biopython/biopython/issues/1551
                    # Essentially, a new change in biopython prevents you from changing the
                    # id to an already existing id which broke this initial script.
                    # Therefore, we now change the ids to "change_counter" which will never look
                    # like a canonical chainid.
                    structure_after[0][
                        structure_before[0].get_list()[i].id
                    ].id = chain_counter
                    chain_counter += "KK"
                except KeyError:
                    pass
                structure_after[0].get_list()[i].id = (
                    structure_before[0].get_list()[i].id
                )
            if len(residues_before[i]) != len(residues_after[i]):
                raise PDBFixerResIdentifiabilityIssue()

        # When exceeding chainid Z, pdbfixer has discarded it, whereas biopython has not.
        # For simplicity, we just discard it as well and pretend it does not exist.
        # This is a very rare instance and will likely never be a problem unless you
        # are extremely unlucky to work with huge proteins where you care about the
        # truncation.
        except IndexError:
            continue

        counter = 99999  # A large residue number that will never exist in a pdb.
        for res1, res2 in zip(residues_before[i], residues_after[i]):
            assert (
                res1.get_resname().strip() == res2.get_resname().strip()
                or pdbfixer.pdbfixer.substitutions[res1.get_resname()].strip()
                == res2.get_resname().strip()
            )
            if res2.id != res1.id:
                try:
                    # Similar issue as previous hack https://github.com/biopython/biopython/issues/1551
                    structure_after[0][chain.get_id()][res1.id].id = (
                        " ",
                        counter,
                        " ",
                    )
                except KeyError:
                    pass
                res2.id = res1.id
                counter += 1

    return structure_after


def clean_pdb(pdb_input_filename: str, out_dir: str, reduce_executable: str):
    """
    Function to clean pdbs using reduce and pdbfixer. The output file will
    have "_cleaned.pdb" suffix.

    Parameters
    ----------
    pdb_input_filename: str
        PDB filename
    out_dir: str
        Output directory.
    reduce_executable: str
        Path to the reduce executable
    """

    pdbid = pdb_input_filename.split("/")[-1].split(".pdb")[0]

    # Step 1: Add hydrogens using reduce program
    with tempfile.NamedTemporaryFile(mode="wt", delete=True) as temp1:
        first_model = _step_1_reduce(
            reduce_executable, pdb_input_filename, pdbid, temp1
        )
        # Step 2: NonHetSelector filter
        with tempfile.NamedTemporaryFile(mode="wt", delete=True) as temp2:
            PDBIO.set_structure(first_model)
            PDBIO.save(temp2, select=NonHetSelector())
            temp2.flush()
            first_model = PDB_PARSER.get_structure(temp2.name, temp2.name)[0]

            # Step 3: Replace altloc chars to " " and use pdbfixer
            with tempfile.NamedTemporaryFile(mode="wt", delete=True) as temp3:
                temp3, fixer = _step_3_pdbfixer(first_model, temp3)

                # Step 4: Correct for pdbfixer not preserving insertion codes
                with tempfile.NamedTemporaryFile(mode="wt", delete=True) as temp4:
                    structure_after = _step_4_fix_numbering(fixer, temp3, temp4)
                    with open(
                        os.path.join(out_dir, pdbid + "_clean.pdb"), "w"
                    ) as outpdb:
                        PDBIO.set_structure(structure_after[0])
                        PDBIO.save(outpdb)


if __name__ == "__main__":
    t0 = time.time()

    # Argument Parser
    parser = argparse.ArgumentParser()
    parser.add_argument("--pdb_file_in", type=str)
    parser.add_argument("--out_dir", type=str)
    parser.add_argument("--reduce_exe", type=str)

    # Parse arguments
    args_dict = vars(parser.parse_args())
    pdb_input_filename = args_dict["pdb_file_in"]
    out_dir = args_dict["out_dir"]
    reduce_executable = args_dict["reduce_exe"]

    # Clean
    clean_pdb(pdb_input_filename, out_dir, reduce_executable)
    t1 = time.time()
    print(f"Time for cleaning {pdb_input_filename}: {t1-t0}")



from typing import Literal

import numpy as np
from dockstring import load_target

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.registry import register_isolated_function
from poli.core.util.chemistry.string_to_molecule import translate_selfies_to_smiles


class IsolatedDockstringFunction(AbstractIsolatedFunction):
    """
    Black box implementation for the Dockstring problem.

    Dockstring is a simple API for assessing the docking score
    of a small molecule to a given protein [1].

    Parameters
    ----------
    target_name : str
        The name of the target protein.
    string_representation : str, optional
        The string representation of the molecules. Either SMILES or SELFIES.
        Default is SMILES.
    batch_size : int, optional
        The batch size for processing multiple inputs simultaneously, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize the computation, by default False.
    num_workers : int, optional
        The number of workers to use for parallel computation, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.

    Attributes
    ----------
    alphabet : dict
        The mapping of symbols to their corresponding indices in the alphabet.

    Methods
    -------
    _black_box(x, context=None)
        The black box function.

    References
    ----------
    [1] “DOCKSTRING: Easy Molecular Docking Yields Better Benchmarks for Ligand Design.”
        García-Ortegón, Miguel, Gregor N. C. Simm, Austin J. Tripp,
        José Miguel Hernández-Lobato, Andreas Bender, and Sergio Bacallado.
        Journal of Chemical Information and Modeling 62, no. 15 (August 8, 2022): 3486-3502.
        https://doi.org/10.1021/acs.jcim.1c01334.
    """

    def __init__(
        self,
        target_name: str,
        string_representation: Literal["SMILES", "SELFIES"] = "SMILES",
    ):
        """
        Initialize the dockstring black box object.

        Parameters
        ----------
        batch_size : int, optional
            The batch size for processing data, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize the processing, by default False.
        num_workers : int, optional
            The number of workers to use for parallel processing, by default None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        target_name : str
            The name of the target protein.
        string_representation : str
            The string representation of the molecules. Either SMILES or SELFIES.
            Default is SMILES.
        """
        assert (
            target_name is not None
        ), "Missing required keyword argument 'target_name'. "

        self.target_name = target_name
        self.string_representation = string_representation

        self.target = load_target(target_name)

    def __call__(self, x: np.ndarray, context=None) -> np.ndarray:
        """Evaluating the black box function.

        Parameters
        ----------
        x : np.ndarray
            A molecule represented as a string of shape [b, L],
            where b is the batch size and L is the length of the string.
            We expect the elements in a row of x to be the tokens
            of a molecule string representation.
        context : any, optional
            Additional context information for the evaluation. Defaults to None.

        Returns
        -------
        y: np.ndarray
            The output of the black box function.

        Raises
        ------
        AssertionError
            If the input is not a 2D array of strings.
        Exception
            If the docking score cannot be computed.
        """
        assert len(x.shape) == 2, "Expected a 2D array of strings. "
        molecules_as_strings = ["".join(x_i) for x_i in x]

        if self.string_representation == "SELFIES":
            molecules_as_smiles = translate_selfies_to_smiles(molecules_as_strings)
        else:
            molecules_as_smiles = molecules_as_strings

        # Parallelization is handled by the __call__ method of
        # the AbstractBlackBox class.
        scores = []
        for smiles in molecules_as_smiles:
            try:
                score = self.target.dock(smiles)[0]
            except Exception:
                score = np.nan
            scores.append(score)

        # Since our goal is maximization, and scores in dockstring
        # are better if they are lower, we return the negative of
        # the scores.
        return -np.array(scores).reshape(-1, 1)


if __name__ == "__main__":
    # One example of loading up this black box:
    # inner_dockstring_black_box = IsolatedDockstringFunction(
    #     target_name="abl1",
    #     string_representation="SMILES",
    # )

    # TODO: maybe force the user to instanciate.
    register_isolated_function(
        IsolatedDockstringFunction,
        name="dockstring__isolated",
        conda_environment_name="poli__dockstring",
    )



"""
Registers the stability and SASA FoldX black box
and objective factory.

FoldX [1] is a simulator that allows for computing the difference
in free energy between a wildtype protein and a mutated protein.
We pair this with biopython [2] to compute the SASA of the mutated
protein.

[1] Schymkowitz, J., Borg, J., Stricher, F., Nys, R., Rousseau, F.,
    & Serrano, L. (2005). The FoldX web server: an online force field.
    Nucleic acids research, 33(suppl_2), W382-W388.
[2] Cock PA, Antao T, Chang JT, Chapman BA, Cox CJ, Dalke A, Friedberg I,
    Hamelryck T, Kauff F, Wilczynski B and de Hoon MJL (2009) Biopython:
    freely available Python tools for computational molecular biology and
    bioinformatics. Bioinformatics, 25, 1422-1423
"""

import warnings
from pathlib import Path
from typing import List, Union

import numpy as np

from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.problem import Problem
from poli.core.util.seeding import seed_numpy, seed_python
from poli.objective_repository.foldx_stability_and_sasa.register import (
    FoldXStabilityAndSASABlackBox,
)


class RFPFoldXStabilityAndSASAProblemFactory(AbstractProblemFactory):
    """
    Factory class for creating FoldX SASA (Solvent Accessible Surface Area) problems.

    Methods
    -------
    create:
        Creates a problem instance with the specified parameters.
    """

    def create(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        n_starting_points: int = None,
        strict: bool = False,
        experiment_id: str = None,
        tmp_folder: Path = None,
        eager_repair: bool = False,
        verbose: bool = False,
        seed: int = None,
        batch_size: int = 1,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Create a RFPFoldXSASABlackBox object and compute the initial values of wildtypes.

        Parameters
        ----------
        wildtype_pdb_path : Union[Path, List[Path]]
            Path or list of paths to the wildtype PDB files.
        n_starting_points: int, optional
            Size of D_0. Default is all available data.
            The minimum number of sequence is given by the Pareto front of the RFP problem, ie. you cannot have less sequences than that.
        strict: bool, optional
            Enable RuntimeErrors if number of starting sequences different to requested number of sequences.
        experiment_id : str, optional
            Identifier for the experiment.
        tmp_folder : Path, optional
            Path to the temporary folder for intermediate files.
        eager_repair : bool, optional
            Flag indicating whether to perform eager repair.
        verbose : bool, optional
            Flag indicating whether to print the output of FoldX.
        seed : int, optional
            Seed for random number generators. If None is passed,
            the seeding doesn't take place.
        batch_size : int, optional
            Number of samples per batch for parallel computation.
        parallelize : bool, optional
            Flag indicating whether to parallelize the computation.
        num_workers : int, optional
            Number of worker processes for parallel computation.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        Returns
        -------
        problem : Problem
            An instance of the RFP problem, containing the black box and the initial wildtype sequences.

        Raises
        ------
        ValueError
            If wildtype_pdb_path is missing or has an invalid type.
        """
        if seed is not None:
            seed_numpy(seed)
            seed_python(seed)

        if wildtype_pdb_path is None:
            raise ValueError(
                "Missing required argument wildtype_pdb_path. "
                "Did you forget to pass it to create()?"
            )

        if isinstance(wildtype_pdb_path, str):
            wildtype_pdb_path = [Path(wildtype_pdb_path.strip())]
        elif isinstance(wildtype_pdb_path, Path):
            wildtype_pdb_path = [wildtype_pdb_path]
        elif isinstance(wildtype_pdb_path, list):
            if isinstance(wildtype_pdb_path[0], str):
                wildtype_pdb_path = [Path(x.strip()) for x in wildtype_pdb_path]
            elif isinstance(wildtype_pdb_path[0], Path):
                pass
        else:
            raise ValueError(
                f"wildtype_pdb_path must be a string or a Path. Received {type(wildtype_pdb_path)}"
            )
        # By this point, we know that wildtype_pdb_path is a
        # list of Path objects.
        if n_starting_points is None:
            n_starting_points = len(wildtype_pdb_path)

        # For a comparable RFP definition we require the sequences of the Pareto front:
        pareto_sequences_name_pdb_dict = {
            "DsRed.M1": "2VAD",
            "DsRed.T4": "2VAE",
            "mScarlet": "5LK4",
            "AdRed": "6AA7",
            "mRouge": "3NED",
            "RFP630": "3E5V",
        }

        if strict and n_starting_points < len(pareto_sequences_name_pdb_dict):
            raise RuntimeError(
                f"Initial number of sequences too low!\nMinimum size {len(pareto_sequences_name_pdb_dict)} , requested {n_starting_points}"
            )

        remaining_n_starting_points = max(
            n_starting_points - len(pareto_sequences_name_pdb_dict.values()), 0
        )
        # filter minimal required Pareto sequences
        pareto_pdb_files = [
            p
            for p in wildtype_pdb_path
            if any(
                [
                    bool(_pdb.lower() in str(p).lower())
                    for _pdb in pareto_sequences_name_pdb_dict.values()
                ]
            )
        ]
        if len(pareto_pdb_files) != len(pareto_sequences_name_pdb_dict):
            raise RuntimeError(
                f"The provided PDB files list is incomplete!\n Required={','.join(list(pareto_sequences_name_pdb_dict.values()))} provided files={pareto_pdb_files}"
            )

        remaining_wildtype_pdb_files = list(
            set(wildtype_pdb_path) - set(pareto_pdb_files)
        )
        np.random.shuffle(remaining_wildtype_pdb_files)
        remaining_wildtype_pdb_files = remaining_wildtype_pdb_files[
            :remaining_n_starting_points
        ]  # subselect w.r.t. requested number of sequences
        pdb_files_for_black_box: List = pareto_pdb_files + remaining_wildtype_pdb_files

        f = FoldXStabilityAndSASABlackBox(
            wildtype_pdb_path=pdb_files_for_black_box,
            experiment_id=experiment_id,
            tmp_folder=tmp_folder,
            eager_repair=eager_repair,
            verbose=verbose,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
            force_isolation=force_isolation,
        )

        # We need to compute the initial values of all wildtypes
        # in wildtype_pdb_path. For this, we need to specify x0,
        # a vector of wildtype sequences. These are padded to
        # match the maximum length with empty strings.
        wildtype_amino_acids_ = f.wildtype_amino_acids
        longest_wildtype_length = max([len(x) for x in wildtype_amino_acids_])

        wildtype_amino_acids = [
            amino_acids + [""] * (longest_wildtype_length - len(amino_acids))
            for amino_acids in wildtype_amino_acids_
        ]

        x0 = np.array(wildtype_amino_acids).reshape(
            len(pdb_files_for_black_box), longest_wildtype_length
        )

        if n_starting_points is not None and x0.shape[0] != n_starting_points:
            if strict:
                raise RuntimeError(
                    f"Requested number of starting sequences different to loaded!\nRequested n={n_starting_points}, loaded n={x0.shape[0]}"
                )
            else:
                warnings.warn(
                    f"Requested number of starting sequences different to loaded!\nRequested n={n_starting_points}, loaded n={x0.shape[0]}"
                )

        problem = Problem(f, x0)

        return problem



import datetime
import os
import pickle
import random
from pathlib import Path

import numpy as np
import pandas as pd
import pytz
import torch
from Bio.PDB.Polypeptide import one_to_index
from scipy.stats import pearsonr
from torch.utils.data import DataLoader

from .cavity_model import DDGDataset, DDGToTensor, ResidueEnvironmentsDataset, ToTensor
from .PrismData import PrismParser, VariantData
from .visualization import learning_curve_cavity, learning_curve_ds


def train_val_split_cavity(
    parsed_pdb_filenames,
    TRAIN_VAL_SPLIT,
    BATCH_SIZE,
    DEVICE,
):
    """
    Performs training and validation split of ResidueEnvironments.
    Note that we do the split on PDB level not on ResidueEnvironment level due to possible leakage.
    """
    n_train_pdbs = int(len(parsed_pdb_filenames) * TRAIN_VAL_SPLIT)
    filenames_train = parsed_pdb_filenames[:n_train_pdbs]
    filenames_val = parsed_pdb_filenames[n_train_pdbs:]
    to_tensor_transformer = ToTensor(DEVICE)

    dataset_train = ResidueEnvironmentsDataset(
        filenames_train, transformer=to_tensor_transformer
    )
    dataset_val = ResidueEnvironmentsDataset(
        filenames_val, transformer=to_tensor_transformer
    )

    dataloader_train = DataLoader(
        dataset_train,
        batch_size=BATCH_SIZE,
        shuffle=True,
        collate_fn=to_tensor_transformer.collate_cat,
        drop_last=True,
    )

    dataloader_val = DataLoader(
        dataset_val,
        batch_size=BATCH_SIZE,
        shuffle=False,
        collate_fn=to_tensor_transformer.collate_cat,
        drop_last=True,
    )

    print(
        f"Training data set includes {len(filenames_train)} pdbs with "
        f"{len(dataset_train)} environments."
    )
    print(
        f"Validation data set includes {len(filenames_val)} pdbs with "
        f"{len(dataset_val)} environments."
    )

    return dataloader_train, dataset_train, dataloader_val, dataset_val


def train_step(
    batch_x,
    batch_y,
    cavity_model_net,
    optimizer,
    loss_function,
):
    """
    Takes a training step with the cavity model
    """
    cavity_model_net.train()
    optimizer.zero_grad()
    batch_y_pred = cavity_model_net(batch_x)
    loss_batch = loss_function(batch_y_pred, torch.argmax(batch_y, dim=-1))
    loss_batch.backward()
    optimizer.step()
    return (batch_y_pred, loss_batch.detach().cpu().item())


def eval_loop(
    cavity_model_net,
    dataloader_val,
    loss_function,
):
    """
    Performs an eval loop with the cavity model
    """
    # Eval loop. Due to memory, we don't pass the whole eval set to the model
    labels_true_val = []
    labels_pred_val = []
    loss_batch_list_val = []
    cavity_model_net.eval()

    for batch_x_val, batch_y_val in dataloader_val:
        batch_y_pred_val = cavity_model_net(batch_x_val)

        loss_batch_val = loss_function(
            batch_y_pred_val, torch.argmax(batch_y_val, dim=-1)
        )
        loss_batch_list_val.append(loss_batch_val.detach().cpu().item())

        labels_true_val.append(torch.argmax(batch_y_val, dim=-1).detach().cpu().numpy())
        labels_pred_val.append(
            torch.argmax(batch_y_pred_val, dim=-1).detach().cpu().numpy()
        )

    acc_val = np.mean(
        (np.reshape(labels_true_val, -1) == np.reshape(labels_pred_val, -1))
    )
    loss_val = np.mean(loss_batch_list_val)

    return acc_val, loss_val


def train_loop(
    dataloader_train,
    dataloader_val,
    cavity_model_net,
    loss_function,
    optimizer,
    EPOCHS,
    PATIENCE_CUTOFF,
):
    current_best_epoch_idx = -1
    current_best_loss_val = 1e4
    patience = 0
    epoch_idx_to_model_path = {}

    acc_val_list = []
    acc_train_list = []
    loss_train_list = []
    """
    Trains the cavity model in a loop over EPOCHS
    """
    for epoch in range(EPOCHS):
        labels_true = []
        labels_pred = []
        loss_batch_list = []
        cavity_model_net.train()

        for batch_x, batch_y in dataloader_train:
            batch_y_pred, loss_batch = train_step(
                batch_x, batch_y, cavity_model_net, optimizer, loss_function
            )
            loss_batch_list.append(loss_batch)

            labels_true.append(torch.argmax(batch_y, dim=-1).detach().cpu().numpy())
            labels_pred.append(
                torch.argmax(batch_y_pred, dim=-1).detach().cpu().numpy()
            )

        # Train epoch metrics
        acc_train = np.mean(
            (np.reshape(labels_true, -1) == np.reshape(labels_pred, -1))
        )
        loss_train = np.mean(loss_batch_list)

        # Validation epoch metrics
        acc_val, loss_val = eval_loop(cavity_model_net, dataloader_val, loss_function)
        acc_val_list.append(acc_val)
        acc_train_list.append(acc_train)
        loss_train_list.append(loss_train)

        print(
            f"Epoch {epoch:2d}. Train loss: {loss_train:5.3f}. "
            f"Train Acc: {acc_train:4.2f}. Val loss: {loss_val:5.3f}. "
            f"Val Acc {acc_val:4.2f}"
        )

        # Save model
        model_path = f"{os.path.dirname(os.getcwd())}/output/cavity_models/cavity_model_{epoch:02d}.pt"
        epoch_idx_to_model_path[epoch] = model_path
        torch.save(cavity_model_net.state_dict(), model_path)

        # Early stopping
        if loss_val < current_best_loss_val:
            current_best_loss_val = loss_val
            current_best_epoch_idx = epoch
            patience = 0
        else:
            patience += 1
        if patience > PATIENCE_CUTOFF:
            print("Early stopping activated.")
            break

    learning_curve_cavity(acc_val_list, acc_train_list, loss_train_list)
    best_model_path = epoch_idx_to_model_path[current_best_epoch_idx]
    print(
        f"Best epoch idx: {current_best_epoch_idx} with validation loss: "
        f"{current_best_loss_val:5.3f} and model_path: "
        f"{best_model_path}"
    )
    return best_model_path


def populate_dfs_with_resenvs(ddg_data, resenv_dataset):
    """
    Populates ddG dfs with the wt ResidueEnvironment objects.
    """
    print(
        "Dropping data points where residue is not defined in structure "
        "or due to missing parsed pdb file"
    )
    # Add wt residue environments to standard ddg data dataframes

    resenvs_ddg_data = []
    for idx, row in ddg_data.iterrows():
        resenv_key = (
            f"{row['pdbid']}{row['chainid']}_"
            f"{row['variant'][1:-1]}{row['variant'][0]}"
        )
        try:
            resenv = resenv_dataset[resenv_key]
            resenvs_ddg_data.append(resenv)
        except KeyError:
            print(resenv_key + " " "Could not be found in structure data")
            resenvs_ddg_data.append(np.nan)
    ddg_data["resenv"] = resenvs_ddg_data
    n_datapoints_before = ddg_data.shape[0]
    ddg_data.dropna(inplace=True)
    n_datapoints_after = ddg_data.shape[0]
    print(
        f"dropped {n_datapoints_before - n_datapoints_after:4d} / "
        f"{n_datapoints_before:4d} data points from dataset."
    )

    # Load PDB amino acid frequencies used to approximate unfolded states
    pdb_nlfs = -np.log(
        np.load(
            f"{os.path.dirname(os.getcwd())}/data/train/cavity/pdb_frequencies.npz"
        )["frequencies"]
    )

    # Add wt and mt idxs and freqs to df
    ddg_data["wt_idx"] = ddg_data.apply(
        lambda row: one_to_index(row["variant"][0]), axis=1
    )
    ddg_data["mt_idx"] = ddg_data.apply(
        lambda row: one_to_index(row["variant"][-1]), axis=1
    )
    ddg_data["wt_nlf"] = ddg_data.apply(lambda row: pdb_nlfs[row["wt_idx"]], axis=1)
    ddg_data["mt_nlf"] = ddg_data.apply(lambda row: pdb_nlfs[row["mt_idx"]], axis=1)


def remove_disulfides(ddg_data):
    """
    Removes residues known to be involved in disulfide bridges from the data
    """
    disulfide_dict = {
        "6B0N": [
            "C23",
            "C42",
            "C87",
            "C94",
            "C99",
            "C116",
            "C154",
            "C163",
            "C176",
            "C186",
            "C197",
            "C205",
            "C254",
            "C288",
            "C335",
            "C342",
            "C365",
            "C392",
            "C544",
            "C550",
        ],
        "1ZG4": ["C52", "C98"],
        "2YP7": [
            "C7",
            "C45",
            "C57",
            "C69",
            "C90",
            "C132",
            "C270",
            "C274",
            "C298",
            "C453",
            "C460",
            "C464",
        ],
        "6MYA": [
            "C5",
            "C43",
            "C56",
            "C68",
            "C91",
            "C137",
            "C276",
            "C280",
            "C304",
            "C459",
            "C466",
            "C470",
        ],
        "1PS1": ["C126", "C134"],
        "6M17": ["C113", "C121", "C341", "C510", "C522"],
        "2R7E": ["C322", "C241", "C179", "C153", "C598", "C679", "C496", "C522"],
        "4WVP": ["C179", "C122", "C158", "C152", "C42", "C26", "C194", "C169"],
    }
    for pdb in disulfide_dict:
        for res in disulfide_dict[pdb]:
            ddg_data = ddg_data[
                ~((ddg_data["pdbid"] == pdb) & (ddg_data["variant"].str.contains(res)))
            ]

    return ddg_data


def fermi_transform(x):
    """
    Applies a Fermi transformation to values
    """
    alpha = 3.0
    beta = 0.4
    y = 1.0 / (1.0 + np.exp(-beta * (x - alpha)))
    return y


def inverse_fermi_transform(x):
    """
    Applies an inverse Fermi transformation to values
    """
    alpha = 3.0
    beta = 0.4
    EPS = 10.0 ** (-12)
    y = 0.0
    if x == 1.0:
        y = 40.0
    elif x > 0.0 and 1.0 > x:
        y = (alpha * beta - np.log(-1.0 + 1.0 / x + EPS)) / beta
    elif x == 0.0:
        y = -40.0
    return y


def get_ddg_dataloader(ddg_data, data_type, BATCH_SIZE, DEVICE):
    """
    Returns a ddG dataloader based on input data and data type
    """
    # Define ddG data set
    ddg_dataset = DDGDataset(ddg_data, transformer=DDGToTensor(data_type, DEVICE))

    # Define case-dependent data loader parameters
    shuffle = False
    if data_type == "train":
        BATCH_SIZE = BATCH_SIZE
        shuffle = True
        drop_last = True
    elif data_type == "val":
        BATCH_SIZE = BATCH_SIZE
        shuffle = False
        drop_last = True
    elif data_type == "test" or data_type == "pred":
        BATCH_SIZE = 100  # Worst-case max gpu memory on small machine
        shuffle = False
        drop_last = False

    # Define data loader
    ddg_dataloader = DataLoader(
        ddg_dataset,
        batch_size=BATCH_SIZE,
        shuffle=shuffle,
        drop_last=drop_last,
        collate_fn=DDGToTensor(data_type, DEVICE).collate_multi,
    )

    return ddg_dataloader


def train_val_split_ds(ddg_data, BATCH_SIZE, DEVICE):
    """
    Performs training and validation split of ddG dataframe.
    Note that we do the split on PDB level not on ResidueEnvironment level due to possible leakage.
    """
    parsed_pdb_filenames = ddg_data["pdbid"].unique()
    filenames_val = np.array(
        ["2GRN", "3UFJ", "6R5K", "6B0N", "4QO1", "1BRW", "2OCP", "1DO6", "1ZG4", "1ARZ"]
    )
    filenames_train = np.setdiff1d(parsed_pdb_filenames, filenames_val)

    ddg_data_train = ddg_data[ddg_data["pdbid"].isin(filenames_train)]
    ddg_data_val = ddg_data[ddg_data["pdbid"].isin(filenames_val)]

    dataloader_train = get_ddg_dataloader(ddg_data_train, "train", BATCH_SIZE, DEVICE)
    dataloader_val = get_ddg_dataloader(ddg_data_val, "val", BATCH_SIZE, DEVICE)

    print(
        f"Training data set includes {len(filenames_train)} pdbs with "
        f"{len(ddg_data_train)} mutations."
    )
    print(f"Training PDBs are: {filenames_train}.")
    print(
        f"Validation data set includes {len(filenames_val)} pdbs with "
        f"{len(ddg_data_val)} mutations."
    )
    print(f"Validation PDBs are: {filenames_val}.")

    return dataloader_train, dataloader_val


def init_lin_weights(m):
    if type(m) == torch.nn.Linear:
        torch.nn.init.kaiming_uniform_(m.weight)
        m.bias.data.fill_(0.01)


def ds_train_val(
    df_ddg,
    dataloader_train,
    dataloader_val,
    cavity_model_net,
    ds_model_net,
    loss_func,
    optimizer,
    model_idx,
    EPOCHS,
    DEVICE,
):
    """
    Trains the downstream model in a loop over EPOCHS.
    """

    # Set seed
    np.random.seed(model_idx)
    random.seed(model_idx)
    torch.manual_seed(model_idx)
    torch.cuda.manual_seed(model_idx)
    torch.cuda.manual_seed_all(model_idx)

    # Initialize
    train_loss_list = []
    val_loss_list = []
    pearson_r_list = []

    for epoch in range(EPOCHS):
        print(f"Epoch: {epoch+1}/{EPOCHS}")

        # Initialize
        train_loss_batch_list = torch.empty(0, dtype=torch.float32).to(DEVICE)
        val_loss_batch_list = torch.empty(0, dtype=torch.float32).to(DEVICE)
        val_ddg_fermi = torch.empty(0, dtype=torch.float32).to(DEVICE)
        val_ddg_fermi_pred = torch.empty(0, dtype=torch.float32).to(DEVICE)
        val_ddg = torch.empty(0, dtype=torch.float32).to(DEVICE)

        # Train loop
        ds_model_net.train()

        for _, _, _, x_cavity_batch, x_ds_batch, ddg_fermi_batch, _ in dataloader_train:
            # Initialize optimizer
            optimizer.zero_grad()

            # Compute predictions
            cavity_pred_batch = cavity_model_net(x_cavity_batch)
            ddg_fermi_batch_pred = ds_model_net(
                torch.cat((cavity_pred_batch, x_ds_batch), 1)
            )
            loss_batch = loss_func(ddg_fermi_batch_pred, ddg_fermi_batch)
            loss_batch.backward()
            optimizer.step()

            # Append to epoch
            train_loss_batch_list = torch.cat(
                (train_loss_batch_list, loss_batch.detach().reshape(-1))
            )

        # Val loop
        ds_model_net.eval()
        with torch.no_grad():
            for (
                _,
                _,
                _,
                val_x_cavity_batch,
                val_x_ds_batch,
                val_ddg_fermi_batch,
                val_ddg_batch,
            ) in dataloader_val:
                # Compute predictions
                val_cavity_pred_batch = cavity_model_net(val_x_cavity_batch)
                val_ddg_fermi_pred_batch = ds_model_net(
                    torch.cat((val_cavity_pred_batch, val_x_ds_batch), 1)
                )
                val_loss_batch = loss_func(
                    val_ddg_fermi_pred_batch, val_ddg_fermi_batch
                )

                # Append to epoch
                val_loss_batch_list = torch.cat(
                    (val_loss_batch_list, val_loss_batch.reshape(-1))
                )
                val_ddg_fermi = torch.cat(
                    (val_ddg_fermi, val_ddg_fermi_batch.reshape(-1)), 0
                )
                val_ddg_fermi_pred = torch.cat(
                    (val_ddg_fermi_pred, val_ddg_fermi_pred_batch.reshape(-1)), 0
                )
                val_ddg = torch.cat((val_ddg, val_ddg_batch.reshape(-1)), 0)

        # Compute epoch metrics
        train_loss_list.append(train_loss_batch_list.mean().cpu().item())
        val_loss_list.append(val_loss_batch_list.mean().cpu().item())
        val_ddg = val_ddg.detach().cpu()
        val_ddg_fermi_pred = val_ddg_fermi_pred.detach().cpu()
        val_ddg_pred = val_ddg_fermi_pred.apply_(lambda x: inverse_fermi_transform(x))
        pearson_r_list.append(pearsonr(val_ddg_pred.numpy(), val_ddg.numpy())[0])

    # Print learning curve
    learning_curve_ds(pearson_r_list, val_loss_list, train_loss_list, model_idx)

    # Save results and model
    lc_results = (pearson_r_list, val_loss_list, train_loss_list, model_idx)
    with open(
        f"/content/output/ds_models/ds_model_{model_idx}/lc_results.pkl",
        "wb",
    ) as f:
        pickle.dump(lc_results, f)
    torch.save(
        ds_model_net.state_dict(),
        f"/content/output/ds_models/ds_model_{model_idx}/model.pt",
    )


def ds_pred(
    cavity_model_net, ds_model_net, df_total, dataset_key, NUM_ENSEMBLE, DEVICE
):
    """
    Makes predictions using the cavity and downstream model (RaSP model)
    """

    # Make data loader
    dataloader = get_ddg_dataloader(df_total, "pred", None, DEVICE)

    # Make predictions
    pdbid = []
    chainid = []
    variant = []
    ddg_fermi_pred = torch.empty(0, 1, dtype=torch.float32).to(DEVICE)

    with torch.no_grad():
        for (
            pdbid_batch,
            chainid_batch,
            variant_batch,
            x_cavity_batch,
            x_ds_batch,
        ) in dataloader:
            # Initialize
            ddg_fermi_pred_batch_ensemble = torch.empty(len(variant_batch), 0).to(
                DEVICE
            )

            # Load ds models in a load
            HOME_DIR = Path(__file__).resolve().home()
            RASP_DIR = HOME_DIR / ".poli_objectives" / "rasp"
            for i in range(NUM_ENSEMBLE):
                model_idx = i
                ds_model_net.load_state_dict(
                    torch.load(
                        RASP_DIR / "ds_models" / f"ds_model_{model_idx}" / "model.pt",
                        map_location=DEVICE,
                        weights_only=True,
                    )
                )
                ds_model_net.eval()

                # Compute predictions
                cavity_pred_batch = cavity_model_net(x_cavity_batch)
                ddg_fermi_pred_batch = ds_model_net(
                    torch.cat((cavity_pred_batch, x_ds_batch), 1)
                )
                ddg_fermi_pred_batch_ensemble = torch.cat(
                    (ddg_fermi_pred_batch_ensemble, ddg_fermi_pred_batch), 1
                )

            # Take median of ensemble predictions
            ddg_fermi_pred_batch = torch.median(
                ddg_fermi_pred_batch_ensemble, 1, keepdim=True
            )[0]

            # Append to epoch results
            pdbid += pdbid_batch
            chainid += chainid_batch
            variant += variant_batch
            ddg_fermi_pred = torch.cat((ddg_fermi_pred, ddg_fermi_pred_batch), 0)

    # Repack data to df
    df_ml = pd.DataFrame(ddg_fermi_pred.cpu().numpy(), columns=["score_ml_fermi"])
    df_ml["score_ml"] = df_ml["score_ml_fermi"].apply(
        lambda x: inverse_fermi_transform(x)
    )
    df_ml.insert(loc=0, column="pdbid", value=np.array(pdbid))
    df_ml.insert(loc=1, column="chainid", value=np.array(chainid))
    df_ml.insert(loc=2, column="variant", value=np.array(variant))
    return df_ml


def cavity_to_prism(df, pdbid, chainid, seq, prism_file):
    """
    Saves a ddG dataframe to a PRISM-format csv file
    """

    # Get relevant columns
    df_prism = df[["variant", "score_ml_fermi", "score_ml"]]

    # Get current CPH time
    timestamp = datetime.datetime.now(pytz.timezone("Europe/Copenhagen")).strftime(
        "%Y-%m-%d %H:%M"
    )

    metadata = {
        "version": 1,
        "protein": {
            "name": "Unknown",
            "organism": "Unknown",
            "uniprot": "Unknown",
            "sequence": seq,
            "pdb": pdbid,
            "chain": chainid,
        },
        "cavity": {
            "version": 1,
        },
        "columns": {
            "score_ml_fermi": "Normalized cavity model ddG predictions (range [0;1])",
            "score_ml": "Cavity model ddG predictions",
        },
        "created": {f"{timestamp} (CPH time) - lasse.blaabjerg@bio.ku.dk"},
    }

    # Write data
    dataset = VariantData(metadata, df_prism)
    comment = ""
    parser = PrismParser()
    parser.write(prism_file, dataset, comment_lines=comment)


def get_seq_from_variant(df):
    """
    Fetches the wt amino acid sequence from a ddG dataframe
    """

    df["pos"] = df["variant"].str.extract("(\d+)").astype(int)
    df["wt"] = df["variant"].str.extract("(^[a-zA-Z]+)")
    seq = ""
    for i in range(1, max(df["pos"] + 1)):
        if i in df["pos"].unique():
            seq += df["wt"][df["pos"] == i].iloc[0]
        else:
            seq += "X"
    return seq


def compute_pdb_combo_corrs(df, dataset_key):
    """
    Computes correlation coefficients between ddGs from a list of PDBs
    """

    parser = PrismParser()

    df["pos"] = df["variant"].str.extract("(\d+)").astype(int)

    combos = [
        ("6V0P", "4GQB"),
        ("6V0P", "AFAR"),
        ("4GQB", "AFAR"),
        ("6B6U", "6NU5"),
        ("6B6U", "AFPK"),
        ("6NU5", "AFPK"),
        ("4Y08", "4OYN"),
        ("4Y08", "AFFH"),
        ("4OYN", "AFFH"),
    ]

    results = []

    for combo in combos:
        df_pdb_1 = df[df["pdbid"] == combo[0]].reset_index()
        df_pdb_2 = df[df["pdbid"] == combo[1]].reset_index()

        pdbid_1 = df_pdb_1["pdbid"][0]
        chainid_1 = df_pdb_1["chainid"][0]
        seq_1 = get_seq_from_variant(df_pdb_1)
        filename_1 = f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/prism_cavity_pred_{pdbid_1}"
        cavity_to_prism(df_pdb_1, pdbid_1, chainid_1, seq_1, filename_1)
        df_pdb_1_prism = parser.read(filename_1)

        pdbid_2 = df_pdb_2["pdbid"][0]
        chainid_2 = df_pdb_2["chainid"][0]
        seq_2 = get_seq_from_variant(df_pdb_2)
        filename_2 = f"{os.path.dirname(os.getcwd())}/output/{dataset_key}/prism_cavity_pred_{pdbid_2}"
        cavity_to_prism(df_pdb_2, pdbid_2, chainid_2, seq_2, filename_2)
        df_pdb_2_prism = parser.read(filename_2)

        # Merge
        df_total = df_pdb_1_prism.merge([df_pdb_2_prism]).dataframe.dropna()
        x = df_total["score_ml_00"]
        y = df_total["score_ml_01"]

        y = y[((x >= -1) & (x <= 7)) & ((y >= -1) & (y <= 7))]
        x = x[((x >= -1) & (x <= 7)) & ((y >= -1) & (y <= 7))]

        pearson_r = pearsonr(x, y)[0]

        print(
            f"Pearson r of RaSP predictions between {combo[0]} and {combo[1]} is: {pearson_r}"
        )
        results.append((combo[0], combo[1], pearson_r))

    with open(
        f"/content/output/{dataset_key}/combo_corr_results.pkl",
        "wb",
    ) as f:
        pickle.dump(results, f)



"""
This is a suite of tests for the Ehrlich functions
proposed by Stanton et al. (2024).
"""

import numpy as np
import pytest

from poli.objective_repository.ehrlich._construct_feasibility_matrix import (
    _construct_transition_matrix,
)
from poli.repository import EhrlichBlackBox, EhrlichProblemFactory


@pytest.mark.parametrize("size", [3, 5, 8, 10])
@pytest.mark.parametrize("seed", [1, 2, 3, 4])
def test_sparse_matrix_construction_is_ergodic_and_aperiodic(size: int, seed: int):
    sparse_transition_matrix = _construct_transition_matrix(size, seed=seed)

    # Checking with the Perron-Frobenius condition
    assert (
        np.linalg.matrix_power(sparse_transition_matrix, (size - 1) ** 2 + 1) > 0.0
    ).all()


@pytest.mark.parametrize("seed", [1, 4])
@pytest.mark.parametrize("sequence_length", [10, 20, 50, 100])
@pytest.mark.parametrize("motif_length", [2, 3, 4, 5])
@pytest.mark.parametrize("n_motifs", [1, 2, 3, 4])
def test_ehrlich_motifs_and_offsets_are_deterministic(
    seed: int, sequence_length: int, motif_length: int, n_motifs: int
):
    if n_motifs * motif_length > sequence_length:
        pytest.skip(
            "The total length of the motifs is greater than the sequence length."
        )
    ehrlich_1 = EhrlichBlackBox(
        sequence_length=sequence_length,
        motif_length=motif_length,
        n_motifs=n_motifs,
        seed=seed,
    )
    ehrlich_2 = EhrlichBlackBox(
        sequence_length=sequence_length,
        motif_length=motif_length,
        n_motifs=n_motifs,
        seed=seed,
    )

    assert np.all(ehrlich_1.motifs == ehrlich_2.motifs)
    assert np.all(ehrlich_1.offsets == ehrlich_2.offsets)


@pytest.mark.parametrize("sequence_length", [10, 20, 50, 100])
@pytest.mark.parametrize("motif_length", [2, 3, 4, 5])
@pytest.mark.parametrize("n_motifs", [1, 2, 3, 4])
def test_ehrlich_gives_different_motifs_for_different_seeds(
    sequence_length, motif_length, n_motifs
):
    if n_motifs * motif_length > sequence_length:
        pytest.skip(
            "The total length of the motifs is greater than the sequence length."
        )

    ehrlich_1 = EhrlichBlackBox(
        sequence_length=sequence_length,
        motif_length=motif_length,
        n_motifs=n_motifs,
        seed=1,
    )
    ehrlich_2 = EhrlichBlackBox(
        sequence_length=sequence_length,
        motif_length=motif_length,
        n_motifs=n_motifs,
        seed=2,
    )

    assert not np.all(ehrlich_1.motifs == ehrlich_2.motifs)


@pytest.mark.parametrize("sequence_length", [10, 20, 50, 100])
@pytest.mark.parametrize("motif_length", [2, 3, 4, 5])
@pytest.mark.parametrize("n_motifs", [1, 2, 3, 4])
def test_ehrlich_function_produces_optimal_sequences(
    sequence_length: int, motif_length: int, n_motifs: int
):
    if n_motifs * motif_length > sequence_length:
        pytest.skip(
            "The total length of the motifs is greater than the sequence length."
        )

    ehrlich = EhrlichBlackBox(
        sequence_length=sequence_length,
        motif_length=motif_length,
        n_motifs=n_motifs,
        seed=1,
    )

    optimal_sequence = ehrlich.construct_optimal_solution()
    assert ehrlich._is_feasible(optimal_sequence)
    assert ehrlich(optimal_sequence) == 1.0


def test_consistency_of_ehrlich_function_motif_matching():
    ehrlich = EhrlichBlackBox(
        sequence_length=10,
        motif_length=3,
        n_motifs=2,
        quantization=3,
        seed=1,
    )

    one_sequence = np.array(["E"] * 10).reshape(1, 10)
    motif_matches = ehrlich._maximal_motif_matches(
        one_sequence, np.array(["E", "V", "D"]), np.array([0, 1, 3])
    )
    assert motif_matches == 1

    another_sequence = "EVEEEEEEEE"
    another_sequence = np.array(list(another_sequence)).reshape(1, 10)
    assert (
        ehrlich._maximal_motif_matches(
            another_sequence, np.array(["E", "V", "D"]), np.array([0, 1, 3])
        )
        == 2
    )

    yet_another_sequence = "EEEEVEDEEE"
    yet_another_sequence = np.array(list(yet_another_sequence)).reshape(1, 10)
    assert (
        ehrlich._maximal_motif_matches(
            yet_another_sequence, np.array(["E", "V", "D"]), np.array([0, 1, 3])
        )
        == 3
    )


def test_creating_a_problem_with_a_factory():
    problem_factory = EhrlichProblemFactory()

    problem = problem_factory.create(
        sequence_length=10,
        motif_length=3,
        n_motifs=2,
        quantization=3,
        seed=1,
    )

    f, x0 = problem.black_box, problem.x0
    _ = f(x0)


def test_creating_with_create():
    from poli import create

    problem = create(
        name="ehrlich",
        sequence_length=10,
        motif_length=3,
        n_motifs=2,
        quantization=3,
        seed=1,
    )
    f, x0 = problem.black_box, problem.x0
    _ = f(x0)


@pytest.mark.parametrize("seed", [1, 2, 3, 4])
@pytest.mark.parametrize("unfeasible_value", [-1.0, 0.0])
def test_unfeasible_value(seed: int, unfeasible_value: float):
    f = EhrlichBlackBox(
        sequence_length=3,
        motif_length=2,
        n_motifs=1,
        seed=seed,
        return_value_on_unfeasible=unfeasible_value,
        alphabet=["A", "B", "C"],
    )

    unfeasible_pairs = np.where(f.transition_matrix == 0)
    for i, j in zip(*unfeasible_pairs):
        assert (
            f(np.array([[f.alphabet[i], f.alphabet[j], f.alphabet[i]]]))
            == unfeasible_value
        )



"""Registers the Super Mario Bros objective function and factory.

The metric being recorded is the number of
jumps Mario makes in a given level. The goal
is, of course, maximizing it while keeping the
level playable.

"""

from pathlib import Path
from typing import List

import numpy as np

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.objective_repository.super_mario_bros.information import smb_info
from poli.objective_repository.super_mario_bros.simulator import (
    test_level_from_int_array,
)

THIS_DIR = Path(__file__).parent.resolve()

# TODO: download the simulator from the internet
# if it doesn't exist
# FIXME: do this after we remove .jar
# files from the python installation.


class SMBIsolatedLogic(AbstractIsolatedFunction):
    """
    Black box which returns how many jumps were performed
    by Mario in the given level.

    Parameters
    ----------
    batch_size : int, optional
        The batch size for simultaneous execution, by default None.
    parallelize : bool, optional
        Flag indicating whether to parallelize execution, by default False.
    num_workers : int, optional
        The number of workers for parallel execution, by default None.
    evaluation_budget:  int, optional
        The maximum number of function evaluations. Default is infinity.

    Attributes
    ----------
    model : torch.nn.Module
        The model used for the simulation.

    Methods
    -------
    _black_box(self, x, context=None)
        Runs the given input x as a flattened level (14x14)
        through the model and returns the number
        of jumps Mario makes in the level. If the
        level is not solvable, returns np.nan.
    """

    def __init__(
        self,
        alphabet: List[str] = smb_info.alphabet,
        max_time: int = 30,
        visualize: bool = False,
        value_on_unplayable: float = np.nan,
    ):
        self.alphabet = alphabet
        self.alphabet_s_to_i = {s: i for i, s in enumerate(alphabet)}
        self.alphabet_i_to_s = {i: s for i, s in enumerate(alphabet)}
        self.max_time = max_time
        self.visualize = visualize
        self.value_on_unplayable = value_on_unplayable

    def __call__(self, x: np.ndarray, context=None) -> np.ndarray:
        """Computes number of jumps in a given latent code x."""
        assert x.ndim == 2, "x must be a 2D array"
        batch_size, _ = x.shape

        # Converting the level to ints
        x = np.array(
            [[self.alphabet_s_to_i[x_ij] for x_ij in x_i] for x_i in x]
        ).reshape(batch_size, 14, 14)

        # Run the model
        jumps_for_all_levels = []
        for level in x:
            res = test_level_from_int_array(
                level, max_time=self.max_time, visualize=self.visualize
            )

            if not isinstance(res, dict):
                raise ValueError(
                    "Something probably went wrong with the Java simulation "
                    "of the level. It is quite likely you haven't set up a "
                    "virtual screen/frame buffer. Check the docs."
                )

            # Return the number of jumps if the level was
            # solved successfully, else return np.nan
            if res["marioStatus"] == 1:
                jumps = res["jumpActionsPerformed"]
            else:
                jumps = self.value_on_unplayable

            jumps_for_all_levels.append(jumps)

        return np.array(jumps_for_all_levels, dtype=float).reshape(batch_size, 1)


if __name__ == "__main__":
    from poli.core.registry import register_isolated_function

    register_isolated_function(
        SMBIsolatedLogic,
        name="super_mario_bros__isolated",
        conda_environment_name="poli__mario",
    )



import numpy as np
import pytest

from poli import objective_factory

ref_aa_seq = "HPETLVKVKDAEDQLGARVGYIELDLNSGKILESFRPEERFPMMSTFKVLLCGAVLSRVDAGQEQLGRRIHYSQNDLVEYSPVTEKHLTDGMTVRELCSAAITMSDNTAANLLLTTIGGPKELTAFLHNMGDHVTRLDRWNPELNEAIPNDERDTTMPAAMATTLRKLLTGELLTLASRQQLIDWMEADKVAGPLLRSALPAGWFIADKSGAGERGSRGIIAALGPDGKPSRIVVIYTTGSQATMDERNRQIAEIGASLIKHW"


@pytest.mark.poli__rmf
def test_force_isolation_rmf_landscape():
    """
    Test if we can force-register the rmf_landscape problem.
    """
    problem = objective_factory.create(
        name="rmf_landscape",
        wildtype=ref_aa_seq,
        kappa=-100,  # keep noise low
        force_isolation=True,
    )
    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)
    assert np.isclose(np.round(y0), 0.0)
    f.terminate()


@pytest.mark.poli__rmf
def test_rmf_landscape_init():
    problem = objective_factory.create(
        name="rmf_landscape",
        wildtype=ref_aa_seq,
        kappa=-100,
    )
    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)
    assert np.isclose(np.round(y0), 0.0)
    f.terminate()


@pytest.mark.poli__rmf
def test_rmf_landscape_batch_eval():
    problem = objective_factory.create(
        name="rmf_landscape",
        wildtype=ref_aa_seq,
    )
    N = 10
    f, x0 = problem.black_box, problem.x0
    _ = f(x0)
    x_t = []
    seq_b = x0.copy()
    seq_b[0, 1] = "Y"
    x_t = np.vstack([seq_b for _ in range(N)])
    assert x_t.shape[0] == N
    yt = f(x_t)
    assert yt.shape[0] == N
    f.terminate()


@pytest.mark.poli__rmf
@pytest.mark.parametrize("seed", [1, 2, 3])
def test_rmf_seed_consistent(seed: int):
    mutation_seq = list(ref_aa_seq)
    mutation_seq[int(len(mutation_seq) / 2)] = "A"
    mutation_seq[int(len(mutation_seq) / 4)] = "H"
    mutation_seq = np.array(mutation_seq)[None, :]
    problem_a = objective_factory.create(
        name="rmf_landscape",
        wildtype=ref_aa_seq,
        seed=seed,
    )
    problem_b = objective_factory.create(
        name="rmf_landscape", wildtype=ref_aa_seq, seed=seed
    )
    f_a, x0_a = problem_a.black_box, problem_a.x0
    y0_a = f_a(x0_a)
    f_b, x0_b = problem_b.black_box, problem_b.x0
    y0_b = f_b(x0_b)

    y1_a = f_a(mutation_seq)
    y1_b = f_b(mutation_seq)
    # test equalities
    assert all([x_a == x_b for x_a, x_b in zip(x0_a[0], x0_b[0])])
    assert y0_a == y0_b
    assert y1_a == y1_b  # value for mutated sequences equal
    f_a.terminate()
    f_b.terminate()


@pytest.mark.poli__rmf
@pytest.mark.parametrize("n_mutations", [1, 2, 3])
def test_rmf_num_mutations_expected_val(n_mutations: int):
    from scipy.stats import genpareto

    SEED = 1
    mutation_seq = list(ref_aa_seq)
    for m in range(n_mutations):
        mutation_seq[int(len(mutation_seq) / 2) - m] = "Y"
    mutation_seq = np.array(mutation_seq)[None, :]
    problem = objective_factory.create(
        name="rmf_landscape",
        kappa=-100,  # set kappa <0 for sampling values close to zero
        c=1,  # set constant to one s.t. number mutations negative additive
        wildtype=ref_aa_seq,
        seed=SEED,
    )

    f, x0 = problem.black_box, problem.x0
    y0 = f(x0)
    y1 = f(mutation_seq)

    rnd_state = np.random.default_rng(SEED)
    ref_noise_0 = genpareto.rvs(f.kappa, size=1, random_state=rnd_state)
    ref_noise_1 = genpareto.rvs(f.kappa, size=1, random_state=rnd_state)

    # black-box value minus noisy component should be approximately mutational distance if c==1
    assert np.isclose(np.round(y0 - ref_noise_0), 0)
    assert np.isclose(np.round(y1 - ref_noise_1), -n_mutations)
    f.terminate()



"""This module implements the FoldX black box and objective factory.

FoldX [1] is a simulator that allows for computing the difference
in free energy between a wildtype protein and a mutated protein.

References
----------
[1] The FoldX web server: an online force field.
    Nucleic acids research Schymkowitz, J., Borg, J., Stricher,
    F., Nys, R., Rousseau, F., & Serrano, L. (2005). ,
    33(suppl_2), W382-W388.
"""

from multiprocessing import cpu_count
from pathlib import Path
from time import time
from typing import List, Union
from uuid import uuid4

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.black_box_information import BlackBoxInformation
from poli.core.util.proteins.foldx import FoldxInterface
from poli.core.util.proteins.pdb_parsing import (
    parse_pdb_as_residue_strings,
    parse_pdb_as_residues,
)

# This is the folder where all the files
# generated by FoldX will be stored.
# Feel free to change it if you want
# to keep the files somewhere else by
# passing tmp_folder to the black box.
# TODO: what happens if the user is on Windows?
DEFAULT_TMP_PATH = Path("/tmp").resolve()


class FoldxBlackBox(AbstractBlackBox):
    """
    A class representing the FoldxBlackBox, which is used for simulating protein mutations using FoldX.

    Parameters
    -----------
    info : ProblemSetupInformation, required
        The problem setup information object. (default: None)
    batch_size : int, optional
        The batch size for parallelization. (default: None)
    parallelize : bool, optional
        Flag indicating whether to parallelize the simulations. (default: False)
    num_workers : int, optional
        The number of workers for parallelization. (default: None)
    wildtype_pdb_path : Union[Path, List[Path]], required
        The path(s) to the wildtype PDB file(s). (default: None)
    alphabet : List[str], optional
        The list of allowed amino acids. (default: None)
    experiment_id : str, optional
        The experiment ID. (default: None)
    tmp_folder : Path, optional
        The temporary folder path. (default: None)
    eager_repair : bool, optional
        Flag indicating whether to eagerly repair the PDB files. (default: False)
    verbose : bool, optional
        Flag indicating whether we print the output from FoldX. (default: False)

    Attributes
    ----------
    experiment_id : str
        The experiment ID.
    tmp_folder : Path
        The temporary folder path.
    wildtype_pdb_paths : List[Path]
        The list of repaired wildtype PDB file paths.
    wildtype_residues : List[List[Residue]]
        The list of wildtype residues for each PDB file.
    wildtype_amino_acids : List[List[str]]
        The list of wildtype amino acids for each PDB file.
    wildtype_residue_strings : List[str]
        The list of wildtype residue strings for each PDB file.

    Methods
    -------
    create_working_directory() -> Path:
        Creates and returns the working directory path for the black box.

    """

    def __init__(
        self,
        info: BlackBoxInformation = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        wildtype_pdb_path: Union[Path, List[Path]] = None,
        alphabet: List[str] = None,
        experiment_id: str = None,
        tmp_folder: Path = None,
        eager_repair: bool = False,
        verbose: bool = False,
    ):
        """
        Initialize the FoldxBlackBox.

        Parameters
        -----------
        info : ProblemSetupInformation, optional
            The problem setup information object. (default: None)
        batch_size : int, optional
            The batch size for parallelization. (default: None)
        parallelize : bool, optional
            Flag indicating whether to parallelize the simulations. (default: False)
        num_workers : int, optional
            The number of workers for parallelization. (default: None)
        evaluation_budget : int, optional
            The evaluation budget. (default: float('inf'))
        wildtype_pdb_path : Union[Path, List[Path]], optional
            The path(s) to the wildtype PDB file(s). (default: None)
        alphabet : List[str], optional
            The list of allowed amino acids. (default: None)
        experiment_id : str, optional
            The experiment ID. (default: None)
        tmp_folder : Path, optional
            The temporary folder path. (default: None)
        eager_repair : bool, optional
            Flag indicating whether to eagerly repair the PDB files. (default: False)
        verbose : bool, optional
            Flag indicating whether we print the output from FoldX. (default: False)
        """
        # WARNING: notice how the batch-size is set to 1.
        # This is because we only support simulating one
        # mutation at a time.
        # TODO: fix this using parallelization.

        # TODO: assert that wildtype_pdb_file is provided
        assert wildtype_pdb_path is not None, (
            "Missing required argument wildtype_pdb_file. "
            "Did you forget to pass it to create and into the black box?"
        )

        # In the specific case of FoldX black boxes,
        # the default batch size is 1. However, if
        # the user passed parallelize=True, then
        # the batch size is set to the number of
        # workers.
        if batch_size is None:
            if parallelize:
                if num_workers is None:
                    num_workers = cpu_count() // 2
                else:
                    batch_size = num_workers
            else:
                batch_size = 1

        super().__init__(
            info=info,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        self.verbose = verbose

        # Defining the experiment id
        if experiment_id is None:
            experiment_id = f"{int(time())}_{str(uuid4())[:8]}"
        self.experiment_id = experiment_id

        self.tmp_folder = tmp_folder if tmp_folder is not None else DEFAULT_TMP_PATH

        if alphabet is None:
            alphabet = info.alphabet

        if isinstance(wildtype_pdb_path, str):
            wildtype_pdb_path = Path(wildtype_pdb_path.strip())

        if isinstance(wildtype_pdb_path, Path):
            wildtype_pdb_path = [wildtype_pdb_path]

        if isinstance(wildtype_pdb_path, list):
            _wildtype_pdb_path = []
            for pdb_file in wildtype_pdb_path:
                if isinstance(pdb_file, str):
                    pdb_file = Path(pdb_file.strip())
                assert isinstance(
                    pdb_file, Path
                ), f"Expected a Path object or a string, but got {type(pdb_file)}."
                _wildtype_pdb_path.append(pdb_file)

            wildtype_pdb_path = _wildtype_pdb_path

        # At this point, wildtype_pdb_path is a list of Path objects.
        # We need to ensure that these are repaired pdb files.
        # We do this by creating a temporary folder and repairing
        # the pdbs there.
        if eager_repair:
            path_for_repairing_pdbs = self.tmp_folder / "foldx_tmp_files_for_repair"
            path_for_repairing_pdbs.mkdir(exist_ok=True, parents=True)
            foldx_interface_for_repairing = FoldxInterface(
                path_for_repairing_pdbs, verbose=self.verbose
            )

            # Re-writing wildtype_pdb_path to be the list of repaired pdb files.
            repaired_wildtype_pdb_files = [
                foldx_interface_for_repairing._repair_if_necessary_and_provide_path(
                    pdb_file
                )
                for pdb_file in wildtype_pdb_path
            ]

            # At this point, wildtype_pdb_path is a list of Path objects.
            self.wildtype_pdb_paths = repaired_wildtype_pdb_files
        else:
            self.wildtype_pdb_paths = wildtype_pdb_path

        self.wildtype_resiudes = [
            parse_pdb_as_residues(pdb_file) for pdb_file in self.wildtype_pdb_paths
        ]

        self.wildtype_amino_acids = [
            parse_pdb_as_residue_strings(pdb_file)
            for pdb_file in self.wildtype_pdb_paths
        ]

        self.wildtype_residue_strings = [
            "".join(amino_acids) for amino_acids in self.wildtype_amino_acids
        ]

    def create_working_directory(self) -> Path:
        """
        Create and return the working directory path for the black box.

        Returns
        --------
        Path
            The path to the working directory.
        """
        sub_experiment_id = str(uuid4())[:8]

        working_dir = (
            self.tmp_folder / "foldx_tmp_files" / self.experiment_id / sub_experiment_id
        )
        working_dir.mkdir(exist_ok=True, parents=True)

        return working_dir



from .foldx_supervised_data_packages import RFPFoldXStabilitySupervisedDataPackage
from .rasp_supervised_data_packages import RFPRaspSupervisedDataPackage

__all__ = [
    "RFPRaspSupervisedDataPackage",
    "RFPFoldXStabilitySupervisedDataPackage",
]



"""
This script contains a data package that is frequently used in
small molecule optimization: sampling random molecules from Zinc250k.
"""

from __future__ import annotations

from pathlib import Path
from typing import Callable, Literal

import numpy as np

from poli.core.data_package import DataPackage
from poli.core.util.chemistry.string_to_molecule import translate_smiles_to_selfies


class RandomMoleculesDataPackage(DataPackage):
    """
    Returns a supervised data package with random molecules from Zinc250k.

    We subsampled 5000 smiles from Zinc250k and stored them in a numpy file,
    and this data package samples n_molecules from this set.

    Parameters
    ----------
    string_representation : Literal["SMILES", "SELFIES"]
        The string representation of the molecules.
    n_molecules : int, optional
        The number of molecules to sample from the dataset, by default 10.
    seed : int, optional
        The seed for the random number generator, by default None.
        If provided, we seed numpy random number generator with this seed.
    tokenize_with : Callable[[str], list[str]], optional
        A function that tokenizes the molecules, by default None.
        If provided, we tokenize the molecules with this function.
    """

    def __init__(
        self,
        string_representation: Literal["SMILES", "SELFIES"],
        n_molecules: int = 10,
        seed: int | None = None,
        tokenize_with: Callable[[str], list[str]] = None,
    ):
        assert (
            n_molecules <= 5000
        ), "This data package has been implemented for up to 5000 random molecules."
        CHEMISTRY_DATA_PACKAGES_DIR = Path(__file__).parent
        five_thousand_molecules = np.load(
            CHEMISTRY_DATA_PACKAGES_DIR / "five_thousand_smiles.npz",
            allow_pickle=True,
        )["x"]

        if string_representation.upper() == "SELFIES":
            five_thousand_molecules_ = translate_smiles_to_selfies(
                five_thousand_molecules,
                strict=True,
            )
            five_thousand_molecules = np.array(five_thousand_molecules_)

        if seed is not None:
            np.random.seed(seed)

        unsupervised_data = np.random.choice(
            five_thousand_molecules, (n_molecules,), replace=False
        )
        supervised_data = None

        if tokenize_with is not None:
            unsupervised_data = np.array(
                [tokenize_with(mol) for mol in unsupervised_data if mol is not None]
            )

        super().__init__(unsupervised_data, supervised_data)



from pathlib import Path

import numpy as np
import pandas as pd

from poli.core.abstract_isolated_function import AbstractIsolatedFunction


class GFPSelectIsolatedLogic(AbstractIsolatedFunction):
    def __init__(
        self,
        seed: int = None,
    ):
        gfp_df_path = Path(__file__).parent.resolve() / "assets" / "gfp_data.csv"
        self.seed = seed
        self.gfp_lookup_df = pd.read_csv(gfp_df_path)[
            ["medianBrightness", "aaSequence"]
        ]

        randomized_df = self.gfp_lookup_df.sample(
            frac=1, random_state=seed
        ).reset_index()
        # create 2D array for blackbox evaluation
        x0 = np.array([list(_s) for _s in randomized_df.aaSequence.to_numpy()])

        self.x0 = x0

    def __call__(self, x: np.array, context=None) -> np.ndarray:
        """
        x is string sequence which we look-up in avilable df, return median Brightness
        """
        if isinstance(x, np.ndarray):
            _arr = x.copy()
            x = ["".join(_seq) for _seq in _arr]
        ys = []
        for _x in x:
            seq_subsets = self.gfp_lookup_df[
                self.gfp_lookup_df.aaSequence.str.lower() == _x.lower()
            ]
            # multiple matches possible, shuffle and return one:
            candidate = seq_subsets.sample(n=1, random_state=self.seed)
            ys.append(candidate.medianBrightness)
        return np.array(ys)


if __name__ == "__main__":
    from poli.core.registry import register_isolated_function

    register_isolated_function(
        GFPSelectIsolatedLogic,
        name="gfp_select__isolated",
        conda_environment_name="poli__protein_cbas",
    )



import numpy as np

from poli.core.black_box_information import BlackBoxInformation

sa_tdc_info = BlackBoxInformation(
    name="sa_tdc",
    max_sequence_length=np.inf,
    aligned=False,
    fixed_length=False,
    deterministic=True,  # ?
    alphabet=None,  # TODO: add alphabet once we settle for one for SMLIES/SELFIES.
    log_transform_recommended=False,
    discrete=True,
    padding_token="",
)



"""
Implements a data package, containing unsupervised and
supervised data.
"""

from __future__ import annotations

import numpy as np


class DataPackage:
    def __init__(
        self,
        unsupervised_data: np.ndarray | None,
        supervised_data: tuple[np.ndarray, np.ndarray] | None,
    ):
        self.unsupervised_data = unsupervised_data
        self.supervised_data = supervised_data

    def __repr__(self) -> str:
        unsupervised_data = (
            self.unsupervised_data.shape if self.unsupervised_data is not None else None
        )
        supervised_data = (
            self.supervised_data[0].shape if self.supervised_data is not None else None
        )
        return f"DataPackage(unsupervised_data={unsupervised_data}, supervised_data={supervised_data})"

    def __str__(self) -> str:
        return self.__repr__()



import os.path

import numpy as np
import pandas as pd

from .information import AA

AA_IDX = {AA[i]: i for i in range(len(AA))}

BLOSUM = np.array(
    [
        [
            3.9029,
            0.6127,
            0.5883,
            0.5446,
            0.8680,
            0.7568,
            0.7413,
            1.0569,
            0.5694,
            0.6325,
            0.6019,
            0.7754,
            0.7232,
            0.4649,
            0.7541,
            1.4721,
            0.9844,
            0.4165,
            0.5426,
            0.9365,
        ],
        [
            0.6127,
            6.6656,
            0.8586,
            0.5732,
            0.3089,
            1.4058,
            0.9608,
            0.4500,
            0.9170,
            0.3548,
            0.4739,
            2.0768,
            0.6226,
            0.3807,
            0.4815,
            0.7672,
            0.6778,
            0.3951,
            0.5560,
            0.4201,
        ],
        [
            0.5883,
            0.8586,
            7.0941,
            1.5539,
            0.3978,
            1.0006,
            0.9113,
            0.8637,
            1.2220,
            0.3279,
            0.3100,
            0.9398,
            0.4745,
            0.3543,
            0.4999,
            1.2315,
            0.9842,
            0.2778,
            0.4860,
            0.3690,
        ],
        [
            0.5446,
            0.5732,
            1.5539,
            7.3979,
            0.3015,
            0.8971,
            1.6878,
            0.6343,
            0.6786,
            0.3390,
            0.2866,
            0.7841,
            0.3465,
            0.2990,
            0.5987,
            0.9135,
            0.6948,
            0.2321,
            0.3457,
            0.3365,
        ],
        [
            0.8680,
            0.3089,
            0.3978,
            0.3015,
            19.5766,
            0.3658,
            0.2859,
            0.4204,
            0.3550,
            0.6535,
            0.6423,
            0.3491,
            0.6114,
            0.4390,
            0.3796,
            0.7384,
            0.7406,
            0.4500,
            0.4342,
            0.7558,
        ],
        [
            0.7568,
            1.4058,
            1.0006,
            0.8971,
            0.3658,
            6.2444,
            1.9017,
            0.5386,
            1.1680,
            0.3829,
            0.4773,
            1.5543,
            0.8643,
            0.3340,
            0.6413,
            0.9656,
            0.7913,
            0.5094,
            0.6111,
            0.4668,
        ],
        [
            0.7413,
            0.9608,
            0.9113,
            1.6878,
            0.2859,
            1.9017,
            5.4695,
            0.4813,
            0.9600,
            0.3305,
            0.3729,
            1.3083,
            0.5003,
            0.3307,
            0.6792,
            0.9504,
            0.7414,
            0.3743,
            0.4965,
            0.4289,
        ],
        [
            1.0569,
            0.4500,
            0.8637,
            0.6343,
            0.4204,
            0.5386,
            0.4813,
            6.8763,
            0.4930,
            0.2750,
            0.2845,
            0.5889,
            0.3955,
            0.3406,
            0.4774,
            0.9036,
            0.5793,
            0.4217,
            0.3487,
            0.3370,
        ],
        [
            0.5694,
            0.9170,
            1.2220,
            0.6786,
            0.3550,
            1.1680,
            0.9600,
            0.4930,
            13.5060,
            0.3263,
            0.3807,
            0.7789,
            0.5841,
            0.6520,
            0.4729,
            0.7367,
            0.5575,
            0.4441,
            1.7979,
            0.3394,
        ],
        [
            0.6325,
            0.3548,
            0.3279,
            0.3390,
            0.6535,
            0.3829,
            0.3305,
            0.2750,
            0.3263,
            3.9979,
            1.6944,
            0.3964,
            1.4777,
            0.9458,
            0.3847,
            0.4432,
            0.7798,
            0.4089,
            0.6304,
            2.4175,
        ],
        [
            0.6019,
            0.4739,
            0.3100,
            0.2866,
            0.6423,
            0.4773,
            0.3729,
            0.2845,
            0.3807,
            1.6944,
            3.7966,
            0.4283,
            1.9943,
            1.1546,
            0.3711,
            0.4289,
            0.6603,
            0.5680,
            0.6921,
            1.3142,
        ],
        [
            0.7754,
            2.0768,
            0.9398,
            0.7841,
            0.3491,
            1.5543,
            1.3083,
            0.5889,
            0.7789,
            0.3964,
            0.4283,
            4.7643,
            0.6253,
            0.3440,
            0.7038,
            0.9319,
            0.7929,
            0.3589,
            0.5322,
            0.4565,
        ],
        [
            0.7232,
            0.6226,
            0.4745,
            0.3465,
            0.6114,
            0.8643,
            0.5003,
            0.3955,
            0.5841,
            1.4777,
            1.9943,
            0.6253,
            6.4815,
            1.0044,
            0.4239,
            0.5986,
            0.7938,
            0.6103,
            0.7084,
            1.2689,
        ],
        [
            0.4649,
            0.3807,
            0.3543,
            0.2990,
            0.4390,
            0.3340,
            0.3307,
            0.3406,
            0.6520,
            0.9458,
            1.1546,
            0.3440,
            1.0044,
            8.1288,
            0.2874,
            0.4400,
            0.4817,
            1.3744,
            2.7694,
            0.7451,
        ],
        [
            0.7541,
            0.4815,
            0.4999,
            0.5987,
            0.3796,
            0.6413,
            0.6792,
            0.4774,
            0.4729,
            0.3847,
            0.3711,
            0.7038,
            0.4239,
            0.2874,
            12.8375,
            0.7555,
            0.6889,
            0.2818,
            0.3635,
            0.4431,
        ],
        [
            1.4721,
            0.7672,
            1.2315,
            0.9135,
            0.7384,
            0.9656,
            0.9504,
            0.9036,
            0.7367,
            0.4432,
            0.4289,
            0.9319,
            0.5986,
            0.4400,
            0.7555,
            3.8428,
            1.6139,
            0.3853,
            0.5575,
            0.5652,
        ],
        [
            0.9844,
            0.6778,
            0.9842,
            0.6948,
            0.7406,
            0.7913,
            0.7414,
            0.5793,
            0.5575,
            0.7798,
            0.6603,
            0.7929,
            0.7938,
            0.4817,
            0.6889,
            1.6139,
            4.8321,
            0.4309,
            0.5732,
            0.9809,
        ],
        [
            0.4165,
            0.3951,
            0.2778,
            0.2321,
            0.4500,
            0.5094,
            0.3743,
            0.4217,
            0.4441,
            0.4089,
            0.5680,
            0.3589,
            0.6103,
            1.3744,
            0.2818,
            0.3853,
            0.4309,
            38.1078,
            2.1098,
            0.3745,
        ],
        [
            0.5426,
            0.5560,
            0.4860,
            0.3457,
            0.4342,
            0.6111,
            0.4965,
            0.3487,
            1.7979,
            0.6304,
            0.6921,
            0.5322,
            0.7084,
            2.7694,
            0.3635,
            0.5575,
            0.5732,
            2.1098,
            9.8322,
            0.6580,
        ],
        [
            0.9365,
            0.4201,
            0.3690,
            0.3365,
            0.7558,
            0.4668,
            0.4289,
            0.3370,
            0.3394,
            2.4175,
            1.3142,
            0.4565,
            1.2689,
            0.7451,
            0.4431,
            0.5652,
            0.9809,
            0.3745,
            0.6580,
            3.6922,
        ],
    ]
)


def get_gfp_base_seq():
    """Returns the wild type GFP sequence"""
    f = open(
        os.path.join(
            os.path.dirname(__file__), "ref", "data", "avGFP_reference_sequence.fa"
        )
    )
    lines = f.readlines()
    f.close()
    seq = lines[1].strip()
    return seq


def read_gfp_data(path=None, df_save_file=None):
    """Reads the GFP brightness data in a pandas DataFrame"""
    if path is None:
        path = "../data/nucleotide_genotypes_to_brightness.tsv"
    f = open(path)
    base_seq = get_base_seq()
    mod = list(base_seq)
    mod[191] = "A"
    base_seq = "".join(mod)
    data = []
    i = 0
    cols = None
    for line in f:
        if i == 0:
            cols = line.strip().split("\t")
            cols = ["nucSequence", "numNucMutations", "numAAMutations"] + cols[2:]
            i += 1
            continue
        else:
            split = line.split("\t")
            mutations = split[0].split(":")
            seq, n_nuc_mut = convert_mutations_to_sequence(base_seq, mutations)
            n_aa_mut = len([m for m in split[1].split(":") if m != ""])
            if split[-1].strip() == "":
                std = 0
            else:
                std = float(split[-1].strip())
            data.append([seq, n_nuc_mut, n_aa_mut, int(split[2]), float(split[3]), std])
            i += 1
    df = pd.DataFrame(data, columns=cols)
    if df_save_file is not None:
        df.to_csv(df_save_file)
    return df


def convert_mutations_to_sequence(base_seq, mutations):
    """
    Given the wild type sequence and a formatted mtuation string, returns
    the mutated sequence
    """
    new_seq = list(base_seq)
    n = 0
    for m in mutations:
        if len(m) == 0:
            continue
        style = m[0]
        assert style == "S"
        nuc1 = m[1]
        nuc2 = m[-1]
        pos = int(m[2:-1])
        assert new_seq[pos] == nuc1
        new_seq[pos] = nuc2
        n += 1
    new_seq = "".join(new_seq)
    return new_seq, n


def one_hot_encode_dna(dna_str, pad=None, base_order="ATCG"):
    """Convert length M string into M x 4 tokenized array"""
    dna_str = dna_str.upper()
    if pad is not None:
        M = pad
    else:
        M = len(dna_str)
    dna_arr = np.zeros((M, 4))
    for i in range(len(dna_str)):
        idx = base_order.index(dna_str[i])
        dna_arr[i, idx] = 1
    return dna_arr


def one_hot_encode_aa_array(X_aa: np.ndarray) -> np.ndarray:
    """
    OneHot encodes array: (batch_size, L) -> (batch_size, L, alphabet_size)
    """
    oh_arr = np.zeros((X_aa.shape[0], X_aa.shape[1], len(AA)), dtype=int)
    for i in range(X_aa.shape[0]):
        for j in range(X_aa.shape[1]):
            oh_arr[i, j, AA_IDX[X_aa[i, j]]] = 1
    # test if first and last entry match
    np.testing.assert_array_equal(
        np.array([AA[idx] for idx in np.argmax(oh_arr[0], axis=1)]), X_aa[0]
    )
    np.testing.assert_array_equal(
        np.array([AA[idx] for idx in np.argmax(oh_arr[-1], axis=1)]), X_aa[-1]
    )
    return oh_arr


def one_hot_encode_aa(aa_str, pad=None):
    """Returns a one hot encoded amino acid sequence"""
    M = len(aa_str)
    aa_arr = np.zeros((M, 20), dtype=int)
    for i in range(M):
        aa_arr[i, AA_IDX[aa_str[i]]] = 1
    return aa_arr


def convert_aas_to_idx_array(X_aa):
    """
    Converts a list of amino acid sequences into an array of amino acid indices from AA_IDX.
    E.g. the amino sequence ARN becomes a row containing [1, 2, 3]
    """
    N = len(X_aa)
    M = len(X_aa[0])
    X_aa_idx = np.zeros((N, M), dtype=int)
    for i in range(N):
        for j in range(M):
            X_aa_idx[i, j] = AA_IDX[X_aa[i][j]]

    return X_aa_idx


def convert_idx_array_to_aas(X_aa):
    """Converts an array containing indices of amino acids into the
    corresponding string amino acid sequences. E.g. a row with [1, 2, 3]
    becomes the sequence ARN."""
    N = len(X_aa)
    M = len(X_aa[0])
    X_aa_str = [["A"] * M] * N
    for i in range(N):
        for j in range(M):
            X_aa_str[i][j] = AA[X_aa[i, j]]
        X_aa_str[i] = "".join(X_aa_str[i])
    return X_aa_str


def get_argmax(Xt_p):
    """Given a categorical probability distribution specifying the probability
    of amino acids at each position in a sequence, returns the most probable sequence"""
    Xt_argmax = np.zeros_like(Xt_p)
    Xt_argmax[
        np.arange(Xt_p.shape[0]).reshape(Xt_p.shape[0], 1),
        np.arange(Xt_p.shape[1]).reshape(1, Xt_p.shape[1]),
        np.argmax(Xt_p, axis=-1),
    ] = 1
    return Xt_argmax


def get_samples(Xt_p):
    """Samples from a categorical probability distribution specifying the probability
    of amino acids at each position in a sequence"""
    Xt_sampled = np.zeros_like(Xt_p)
    for i in range(Xt_p.shape[0]):
        for j in range(Xt_p.shape[1]):
            p = Xt_p[i, j]
            k = np.random.choice(range(len(p)), p=p)
            Xt_sampled[i, j, k] = 1.0
    return Xt_sampled


def get_balaji_predictions(preds, Xt):
    """Given a set of predictors built according to the methods in
    the Balaji Lakshminarayanan paper 'Simple and scalable predictive
    uncertainty estimation using deep ensembles' (2017), returns the mean and
    variance of the total prediction."""
    M = len(preds)
    N = Xt.shape[0]
    means = np.zeros((M, N))
    variances = np.zeros((M, N))
    for m in range(M):
        y_pred = preds[m].predict(Xt)
        means[m, :] = y_pred[:, 0]
        variances[m, :] = np.log(1 + np.exp(y_pred[:, 1])) + 1e-6
    mu_star = np.mean(means, axis=0)
    var_star = (1 / M) * (
        np.sum(variances, axis=0) + np.sum(means**2, axis=0)
    ) - mu_star**2
    return mu_star, var_star


def partition_data(
    X, y, percentile=40, train_size=1000, random_state=1, return_test=False
):
    """Partition a (X, y) data set by a percentile of the y values"""
    np.random.seed(random_state)
    assert percentile * 0.01 * len(y) >= train_size
    y_percentile = np.percentile(y, percentile)
    idx = np.where(y < y_percentile)[0]
    rand_idx = np.random.choice(idx, size=train_size, replace=False)
    X_train = X[rand_idx]
    y_train = y[rand_idx]
    if return_test:
        test_idx = [i for i in idx if i not in rand_idx]
        X_test = X[test_idx]
        y_test = y[test_idx]
        return X_train, y_train, X_test, y_test
    else:
        return X_train, y_train


def get_experimental_X_y(
    random_state=1, train_size=5000, return_test=False, return_all=False
):
    """For the GFP testing experiments. Loads the ground truth data (i.e. predictions from the GP model
    we use as the ground truth), partitions it such that we only observe values below the 20th percentile
    and adds measurement noise"""
    prefix = os.path.join(os.path.dirname(os.path.abspath(__file__)), "ref")
    df = pd.read_csv(prefix + "/data/gfp_data.csv")
    X, _ = get_gfp_X_y_aa(df, large_only=True, ignore_stops=True)
    y_gt = np.load(prefix + "/data/gfp_gt_evals.npy")
    if return_test:
        X_train, gt_train, X_test, gt_test = partition_data(
            X,
            y_gt,
            percentile=20,
            train_size=train_size,
            random_state=random_state,
            return_test=return_test,
        )
        np.random.seed(random_state)
        gt_var = 0.01
        y_train = gt_train + np.random.randn(*gt_train.shape) * gt_var
        y_test = gt_test + np.random.randn(*gt_test.shape) * gt_var
        return X_train, y_train, gt_train, X_test, y_test, gt_test
    else:
        X_train, gt_train = partition_data(
            X,
            y_gt,
            percentile=20,
            train_size=train_size,
            random_state=random_state,
            return_test=return_test,
        )
        np.random.seed(random_state)
        gt_var = 0.01
        y_train = gt_train + np.random.randn(*gt_train.shape) * gt_var
        return X_train, y_train, gt_train


def get_gfp_X_y_aa(
    data_df: pd.DataFrame,
    functional_only: bool = False,
    ignore_stops: bool = True,
    return_str: bool = False,
):
    """
    Converts the raw GFP data to a set of X and y values that are ready to use
    in a model
    """
    if functional_only:
        idx = data_df.loc[
            (data_df["medianBrightness"] > data_df["medianBrightness"].mean())
        ].index
    else:
        idx = data_df.index
    data_df = data_df.loc[idx]

    if ignore_stops:
        idx = data_df.loc[~data_df["aaSequence"].str.contains("!")].index
    data_df = data_df.loc[idx]
    seqs = data_df["aaSequence"]

    M = len(seqs[0])
    N = len(seqs)
    X = np.zeros((N, M, 20))
    j = 0
    for i in idx:
        X[j] = one_hot_encode_aa(seqs[i])
        j += 1
    y_raw = np.array(data_df["medianBrightness"][idx])
    y = y_raw
    if return_str:
        return X, y, list(data_df["aaSequence"])
    else:
        return X, y



from typing import Any

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util.inter_process_communication.process_wrapper import ProcessWrapper


class ExternalFunction(AbstractIsolatedFunction):
    """An external version of the black-box function to be instantiated in isolated processes."""

    def __init__(self, process_wrapper: ProcessWrapper):
        """
        Initialize the ExternalBlackBox object.

        Parameters
        ----------
        process_wrapper : ProcessWrapper
            The process wrapper to communicate with the isolated function process.
        """
        # We don't need to pass the kwargs to the parent class,
        # because we overwrite the __getattr__ method to communicate
        # with the isolated objective process.
        super().__init__()
        self.process_wrapper = process_wrapper

    def __call__(self, x, context=None):
        """
        Evaluates the black-box function.

        In this external black-box, the evaluation is done by sending a message
        to the objective process and waiting for the response. The interprocess
        communication is handled by the ProcessWrapper class, which maintains
        an isolated process in which a black-box function runs with, potentially,
        a completely different python executable (e.g. inside another conda
        environment).

        Parameters
        ----------
        x : np.ndarray
            The input data points.
        context : object
            Additional context for the observation.

        Returns
        -------
        y : np.ndarray
            The output data points.
        """
        self.process_wrapper.send(["QUERY", x, context])
        msg_type, *val = self.process_wrapper.recv()
        if msg_type == "EXCEPTION":
            e, traceback_ = val
            print(traceback_)
            raise e
        elif msg_type == "QUERY":
            y = val[0]

            return y
        else:
            raise ValueError(
                f"Internal error: received {msg_type} when expecting QUERY or EXCEPTION"
            )

    def terminate(self):
        """Terminates the external black box."""
        # terminate objective process
        if self.process_wrapper is not None:
            try:
                self.process_wrapper.send(["QUIT", None])
                self.process_wrapper.close()  # clean up connection
            except AttributeError:
                # This means that the process has already been terminated
                pass
            self.process_wrapper = None

    def __getattr__(self, __name: str) -> Any:
        """Gets an attribute from the underlying isolated function.

        Asks for the attribute of the underlying
        isolated function by sending a message
        to the process w. the msg_type "ATTRIBUTE".

        Parameters
        ----------
        __name : str
            The name of the attribute.

        Returns
        -------
        attribute : Any
            The attribute of the underlying black-box function.
        """
        self.process_wrapper.send(["IS_METHOD", __name])
        msg_type, *msg = self.process_wrapper.recv()
        if msg_type == "EXCEPTION":
            e, traceback_ = msg
            print(traceback_)
            raise e
        else:
            assert msg_type == "IS_METHOD"
            is_method = msg[0]

        if is_method:
            return lambda *args, **kwargs: self._method_call(__name, *args, **kwargs)
        else:
            self.process_wrapper.send(["ATTRIBUTE", __name])
            msg_type, *msg = self.process_wrapper.recv()
            if msg_type == "EXCEPTION":
                e, traceback_ = msg
                print(traceback_)
                raise e
            else:
                assert msg_type == "ATTRIBUTE"
                attribute = msg[0]
                return attribute

    def _method_call(self, method_name: str, *args, **kwargs) -> Any:
        """Calls a method of the underlying isolated function.

        Asks for the method of the underlying
        isolated function by sending a message
        to the process w. the msg_type "METHOD".

        Parameters
        ----------
        method_name : str
            The name of the method.
        """
        self.process_wrapper.send(["METHOD", method_name, args, kwargs])
        msg_type, *msg = self.process_wrapper.recv()
        if msg_type == "EXCEPTION":
            e, traceback_ = msg
            print(traceback_)
            raise e
        else:
            assert msg_type == "METHOD"
            return msg[0]

    def __del__(self):
        self.terminate()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.terminate()



"""Tests for the basic promises regarding observers

This module implements tests for
- whether we can define an observer,
- whether the observer properly logs the
  objective calls, and
- whether we can register observers and run
  them in isolated processes using `set_observer`.
"""

import json
from pathlib import Path

import numpy as np

from poli import objective_factory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.registry import register_observer
from poli.core.util.abstract_observer import AbstractObserver

THIS_DIR = Path(__file__).parent.resolve()


class SimpleObserver(AbstractObserver):
    def initialize_observer(
        self,
        problem_setup_info: BlackBoxInformation,
        caller_info: object,
        seed: int,
    ) -> object:
        experiment_id = caller_info["experiment_id"]

        self.experiment_id = experiment_id

        # Creating a local directory for the results
        experiment_path = THIS_DIR / "results" / experiment_id
        experiment_path.mkdir(exist_ok=True, parents=True)

        self.experiment_path = experiment_path
        self.results = []
        # Saving the metadata for this experiment
        metadata = problem_setup_info.as_dict()

        # Adding the information the user wanted to provide
        # (Recall that this caller info gets forwarded
        # from the objective_factory.create function)
        metadata["caller_info"] = caller_info
        metadata["seed"] = seed

        # Saving the metadata
        with open(self.experiment_path / "metadata.json", "w") as f:
            json.dump(metadata, f)

        return None

    def observe(self, x: np.ndarray, y: np.ndarray, context=None) -> None:
        # Appending these results to the results file.
        self.results.append({"x": x.tolist(), "y": y.tolist()})


def test_registering_observer_with_bad_names():
    # TODO[SIMON]: implement
    pass


def test_simple_observer_can_be_defined():
    """Tests whether we can define a simple observer"""
    # Creating a simple observer
    observer = SimpleObserver()

    # Checking whether the observer is an instance of
    # AbstractObserver
    assert isinstance(observer, AbstractObserver)


def test_simple_observer_logs_properly():
    """Tests whether the __call__ of a black box function is properly logged"""
    from poli.core.registry import register_observer

    observer = SimpleObserver()

    register_observer(
        observer=observer,
        conda_environment_location="poli__chem",
        observer_name="simple__",
        set_as_default_observer=False,
    )

    # Creating a black box function
    problem = objective_factory.create(
        name="aloha",
        observer_name="simple__",
        observer_init_info={"experiment_id": "example"},
    )
    f = problem.black_box

    # Evaluating the black box function
    f(np.array([list("MIGUE")]))

    # Checking whether the results were properly logged
    assert f.observer.results == [{"x": [["M", "I", "G", "U", "E"]], "y": [[0.0]]}]
    (f.observer.experiment_path / "metadata.json").unlink()
    f.observer.experiment_path.rmdir()


def test_observer_registration_and_external_instancing():
    """An integration test for the observer registration and external instancing"""
    from poli.core.registry import register_observer

    observer = SimpleObserver()
    register_observer(
        observer=observer.__class__,
        conda_environment_location="poli__chem",
        observer_name="simple__",
        set_as_default_observer=False,
    )

    # Creating a black box function
    problem = objective_factory.create(
        name="aloha",
        observer_name="simple__",
        observer_init_info={"experiment_id": "example"},
    )
    f = problem.black_box

    # Evaluating the black box function
    f(np.array([list("MIGUE")]))

    # Checking whether accessing an unexisting attribute raises an error
    # We do it without pytest to avoid having to install it in
    # the poli__chem environment.
    f.observer  # The same as problem.observer._observer
    try:
        f.observer.unexisting_attribute
    except AttributeError:
        pass

    # Cleaning up (and testing whether we can access attributes
    # of the external observer)
    print(f.observer.experiment_path)
    (f.observer.experiment_path / "metadata.json").unlink()
    f.observer.finish()


def test_multiple_observer_registration():
    observer = SimpleObserver()
    register_observer(
        observer=observer,
        conda_environment_location="poli__chem",
        observer_name="simple__",
        set_as_default_observer=False,
    )

    observer_2 = SimpleObserver()
    register_observer(
        observer=observer_2,
        conda_environment_location="poli__chem",
        observer_name="simple_2__",
        set_as_default_observer=False,
    )

    # Creating a black box function
    problem_1 = objective_factory.create(
        name="aloha",
        observer_name="simple__",
        observer_init_info={"experiment_id": "example"},
    )
    problem_2 = objective_factory.create(
        name="aloha",
        observer_name="simple_2__",
        observer_init_info={"experiment_id": "example_2"},
    )

    # Evaluating the black box function
    problem_1.black_box(np.array([list("MIGUE")]))
    problem_2.black_box(np.array([list("MIGUE")]))

    # Cleaning up (and testing whether we can access attributes
    # of the external observer)
    (problem_1.observer._observer.experiment_path / "metadata.json").unlink()
    (problem_2.observer._observer.experiment_path / "metadata.json").unlink()
    problem_1.observer._observer.finish()
    problem_2.observer._observer.finish()


def test_attaching_an_observer_to_a_black_box():
    from poli.repository import ToyContinuousBlackBox

    f = ToyContinuousBlackBox(
        function_name="ackley_function_01",
        n_dimensions=10,
    )

    observer = SimpleObserver()

    f.set_observer(observer)

    observer.initialize_observer(f.info, {"experiment_id": "attaching"}, seed=0)

    f(np.array([0.0] * 10).reshape(1, 10))

    assert len(observer.results[0]["x"]) == 1



import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.layers import (
    Add,
    Dense,
    Flatten,
    Input,
    Lambda,
    Multiply,
    Reshape,
)
from tensorflow.keras.models import Model
from tensorflow.python.keras import backend as K

tf.compat.v1.disable_v2_behavior()


MAKE_DETERMINISTIC = True


def build_vae(
    latent_dim,
    n_tokens=4,
    seq_length=33,
    enc1_units=50,
    eps_std=1.0,
):
    """Returns a compiled VAE model"""
    model = SimpleVAE(
        input_shape=(
            seq_length,
            n_tokens,
        ),
        latent_dim=latent_dim,
    )

    # set encoder layers:
    model.encoderLayers_ = [
        Dense(units=enc1_units, activation="elu", name="e2"),
    ]

    # set decoder layers:
    model.decoderLayers_ = [
        Dense(units=enc1_units, activation="elu", name="d1"),
        Dense(units=n_tokens * seq_length, name="d3"),
        Reshape((seq_length, n_tokens), name="d4"),
        Dense(units=n_tokens, activation="softmax", name="d5"),
    ]

    # build models:
    kl_scale = K.variable(1.0)
    model.build_encoder()
    model.build_decoder(decode_activation="softmax")
    model.build_vae(epsilon_std=eps_std, kl_scale=kl_scale)

    losses = [summed_categorical_crossentropy, identity_loss]

    model.compile(optimizer="adam", loss=losses)

    return model


def summed_categorical_crossentropy(y_true, y_pred):
    """Negative log likelihood of categorical distribution"""
    return K.sum(K.categorical_crossentropy(y_true, y_pred), axis=-1)


def identity_loss(y_true, y_pred):
    """Returns the predictions"""
    return y_pred


class BaseVAE(object):
    """Base class for Variational Autoencoders implemented in Keras

    The class is designed to connect user-specified encoder and decoder
    models via a Model representing the latent space

    """

    def __init__(self, input_shape, latent_dim, *args, **kwargs):
        self.latentDim_ = latent_dim
        self.inputShape_ = input_shape

        self.encoder_ = None
        self.decoder_ = None

        self.vae_ = None

    def build_encoder(self, *args, **kwargs):
        """Build the encoder network as a keras Model

        The encoder Model must ouput the mean and log variance of
        the latent space embeddings. I.e. this model must output
        mu and Sigma of the latent space distribution:

                    q(z|x) = N(z| mu(x), Sigma(x))

        Sets the value of self.encoder_ to a keras Model

        """

        raise NotImplementedError

    def build_decoder(self, *args, **kwargs):
        """Build the decoder network as a keras Model

        The input to the decoder must have the same shape as the latent
        space and the output must have the same shape as the input to
        the encoder.

        Sets the value of self.decoder_ to a keras Model

        """

        raise NotImplementedError

    def _build_latent_vars(self, mu_z, log_var_z, epsilon_std=1.0, kl_scale=1.0):
        """Build keras variables representing the latent space

        First, calculate the KL divergence from the input mean and log variance
        and add this to the model loss via a KLDivergenceLayer. Then sample an epsilon
        and perform a location-scale transformation to obtain the latent embedding, z.

        Args:
            epsilon_std: standard deviation of p(epsilon)
            kl_scale: weight of KL divergence loss

        Returns
            Variables representing z and epsilon

        """

        # mu_z, log_var_z, kl_batch  = KLDivergenceLayer()([mu_z, log_var_z], scale=kl_scale)
        lmda_func = lambda inputs: -0.5 * K.sum(
            1 + inputs[1] - K.square(inputs[0]) - K.exp(inputs[1]), axis=1
        )

        kl_batch = Lambda(lmda_func, name="kl_calc")([mu_z, log_var_z])
        kl_batch = Reshape((1,), name="kl_reshape")(kl_batch)

        # get standard deviation from log variance:
        sigma_z = Lambda(lambda lv: K.exp(0.5 * lv))(log_var_z)

        if MAKE_DETERMINISTIC:
            eps = Input(tensor=K.zeros_like(mu_z))
        else:
            # re-parametrization trick ( z = mu_z + eps * sigma_z)
            eps = Input(
                tensor=K.random_normal(
                    stddev=epsilon_std, shape=(K.shape(mu_z)[0], self.latentDim_)
                )
            )

        eps_z = Multiply()([sigma_z, eps])  # scale by epsilon sample
        z = Add()([mu_z, eps_z])

        return z, eps, kl_batch

    def _get_decoder_input(self, z, enc_in):
        return z

    def build_vae(self, epsilon_std=1.0, kl_scale=1.0):
        """Build the VAE

        Sets the value of self.vae_ to a keras model

        Args:
            epsilon_std (float): standard deviation of distribution used to sample z via the reparameratization trick
            kl_scale (float or keras.backend.Variable): weight of KL divergence loss

        """

        if self.encoder_ is None:
            raise TypeError("Encoder must be built before calling build_vae")
        if self.decoder_ is None:
            raise TypeError("Decoder must be built before calling build_vae")

        enc_in = self.encoder_.inputs
        mu_z, log_var_z = self.encoder_.outputs
        z, eps, kl_batch = self._build_latent_vars(
            mu_z, log_var_z, epsilon_std=epsilon_std, kl_scale=kl_scale
        )
        dec_in = self._get_decoder_input(z, enc_in)
        x_pred = self.decoder_(dec_in)
        self.vae_ = Model(
            inputs=enc_in + [eps], outputs=[x_pred, kl_batch], name="vae_base"
        )

    def plot_model(self, *args, **kwargs):
        keras.utils.plot_model(self.vae_, *args, **kwargs)

    def compile(self, *args, **kwargs):
        self.vae_.compile(*args, **kwargs)

    def fit(self, *args, **kwargs):
        self.vae_.fit(*args, **kwargs)

    def save_all_weights(self, prefix):
        encoder_file = prefix + "_encoder.h5"
        decoder_file = prefix + "_decoder.h5"
        vae_file = prefix + "_vae.h5"

        self.encoder_.save_weights(encoder_file)
        self.decoder_.save_weights(decoder_file)
        self.vae_.save_weights(vae_file)

    def load_all_weights(self, prefix):
        encoder_file = prefix + "_encoder.h5"
        decoder_file = prefix + "_decoder.h5"
        vae_file = prefix + "_vae.h5"

        self.encoder_.load_weights(encoder_file)
        self.decoder_.load_weights(decoder_file)
        self.vae_.load_weights(vae_file)


class SimpleVAE(BaseVAE):
    """Basic VAE where the encoder and decoder can be constructed from lists of layers"""

    def __init__(self, input_shape, latent_dim, flatten=True, *args, **kwargs):
        super(SimpleVAE, self).__init__(
            input_shape=input_shape, latent_dim=latent_dim, *args, **kwargs
        )
        self.flatten_ = flatten
        self.encoderLayers_ = []
        self.decoderLayers_ = []

    def add_encoder_layer(self, layer):
        """Append a keras Layer to self.encoderLayers_"""
        self.encoderLayers_.append(layer)

    def add_decoder_layer(self, layer):
        """Append a keras Layer to self.decoderLayers_"""
        self.decoderLayers_.append(layer)

    def _build_encoder_inputs(self):
        """BUILD (as opposed to get) the encoder inputs"""
        x = Input(shape=self.inputShape_)
        return [x]

    def _build_decoder_inputs(self):
        z = Input(shape=(self.latentDim_,))
        return z

    def _edit_encoder_inputs(self, enc_in):
        if self.flatten_:
            h = Flatten()(enc_in[0])
        else:
            h = enc_in[0]
        return h

    def _edit_decoder_inputs(self, dec_in):
        return dec_in

    def build_encoder(self):
        """Construct the encoder from list of layers

        After the final layer in self.encoderLayers_, two Dense layers
        are applied to output mu_z and log_var_z

        """

        if len(self.encoderLayers_) == 0:
            raise ValueError("Must add at least one encoder hidden layer")

        enc_in = self._build_encoder_inputs()
        h = self._edit_encoder_inputs(enc_in)
        for hid in self.encoderLayers_:
            h = hid(h)

        mu_z = Dense(self.latentDim_, name="mu_z")(h)
        log_var_z = Dense(self.latentDim_, name="log_var_z")(h)

        self.encoder_ = Model(inputs=enc_in, outputs=[mu_z, log_var_z], name="encoder")

    def build_decoder(self, decode_activation):
        """Construct the decoder from list of layers

        After the final layer in self.decoderLayers_, a Dense layer is
        applied to output the final reconstruction

        Args:
            decode_activation: activation of the final decoding layer

        """

        if len(self.decoderLayers_) == 0:
            raise ValueError("Must add at least one decoder hidden layer")

        dec_in = self._build_decoder_inputs()
        h = self._edit_decoder_inputs(dec_in)
        for hid in self.decoderLayers_:
            h = hid(h)

        x_pred = h
        self.decoder_ = Model(inputs=dec_in, outputs=x_pred, name="decoder")


def build_vae(
    latent_dim,
    n_tokens=4,
    seq_length=33,
    enc1_units=50,
    eps_std=1.0,
):
    """Returns a compiled VAE model"""
    model = SimpleVAE(
        input_shape=(
            seq_length,
            n_tokens,
        ),
        latent_dim=latent_dim,
    )

    # set encoder layers:
    model.encoderLayers_ = [
        Dense(units=enc1_units, activation="elu", name="e2"),
    ]

    # set decoder layers:
    model.decoderLayers_ = [
        Dense(units=enc1_units, activation="elu", name="d1"),
        Dense(units=n_tokens * seq_length, name="d3"),
        Reshape((seq_length, n_tokens), name="d4"),
        Dense(units=n_tokens, activation="softmax", name="d5"),
    ]

    # build models:
    kl_scale = K.variable(1.0)
    model.build_encoder()
    model.build_decoder(decode_activation="softmax")
    model.build_vae(epsilon_std=eps_std, kl_scale=kl_scale)

    losses = [summed_categorical_crossentropy, identity_loss]

    model.compile(optimizer="adam", loss=losses)

    return model



"""This module implements the abstract problem factory."""

from poli.core.problem import Problem


class MetaProblemFactory(type):
    """
    Metaclass for the AbstractProblemFactory class.
    (which allows us to override the __repr__ and __str__ methods)
    """

    def __repr__(cls) -> str:
        return f"<{cls.__name__}()>"

    def __str__(cls) -> str:
        return f"{cls.__name__}"


class AbstractProblemFactory(metaclass=MetaProblemFactory):
    """
    Abstract base class for problem factories.

    This class defines the interface for creating problem instances in poli.

    Methods
    -------
    create:
        Creates a problem instance with the specified parameters.
    """

    def create(
        self,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Returns a blackbox function and initial observations.

        Parameters
        -----------
        seed:  int
            The seed for random number generation. Default is None.
        batch_size:  int
            The batch size for parallel evaluation. Default is None.
        parallelize : bool
            Flag indicating whether to parallelize the evaluation. Default is False.
        num_workers:  int
            The number of workers for parallel evaluation. Default is None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.
        force_isolation: bool, optional
            Whether to force the isolation of the black box. Default
            is False.
        Returns
        --------
        problem: AbstractProblem
            A problem class containing, among other things, the black box,
            initial values x0 and y0, and evaluation budget.

        Raises
        -------
            NotImplementedError: This method is abstract and must be implemented by subclasses.

        """
        raise NotImplementedError(
            "Abstract method create() must be implemented by subclasses."
        )



"""This module tests whether giving black boxes an array of b strings
is equivalent to giving them an array of [b, L] tokens."""

from typing import List

import pytest


# TODO: parametrize by all non-aligned blackboxes
@pytest.mark.parametrize(
    "black_box_name, example_non_flat_input, example_flat_input, kwargs",
    [
        [
            "aloha",
            [["A", "L", "O", "O", "F"], ["A", "L", "O", "H", "A"]],
            [
                "ALOOF",
                "ALOHA",
            ],
            {},
        ],
        [
            "dockstring",
            [
                ["C", ""],
                ["C", "C"],
            ],
            [
                "C",
                "CC",
            ],
            {
                "target_name": "ABL1",
            },
        ],
        # We remove drd3 docking from this check, because
        # it doesn't register out-of-the-box. It needs the user
        # to download a couple of files before using it.
        # [
        #     "drd3_docking",
        #     [
        #         ["C", ""],
        #         ["C", "C"],
        #     ],
        #     [
        #         "C",
        #         "CC",
        #     ],
        #     {},
        # ],
        # TODO: add foldx-related black boxes
        # TODO: Once we can add lambo automatically, we can
        # uncomment this test:
        # [
        #     "penalized_logp_lambo",
        #     [
        #         ["C", ""],
        #         ["C", "C"],
        #     ],
        #     [
        #         "C",
        #         "CC",
        #     ],
        #     {},
        # ],
        # TODO: add rasp
        [
            "rdkit_logp",
            [
                ["C", ""],
                ["C", "C"],
            ],
            [
                "C",
                "CC",
            ],
            {},
        ],
        [
            "rdkit_qed",
            [
                ["C", ""],
                ["C", "C"],
            ],
            [
                "C",
                "CC",
            ],
            {},
        ],
        [
            "sa_tdc",
            [
                ["C", ""],
                ["C", "C"],
            ],
            [
                "C",
                "CC",
            ],
            {},
        ],
    ],
)
def test_passing_array_of_strings(
    black_box_name: str,
    example_non_flat_input: List[List[str]],
    example_flat_input: List[str],
    kwargs: dict,
):
    """This test checks whether passing an array of strings [b,]
    to a black box is equivalent to passing an array of [b, L] tokens.
    """
    import numpy as np

    from poli import create

    problem = create(
        name=black_box_name,
        **kwargs,
    )
    f = problem.black_box

    x_flat = np.array(example_flat_input)
    x_non_flat = np.array(example_non_flat_input)

    assert np.array_equal(f(x_flat), f(x_non_flat), equal_nan=True)



"""
This module implements Rapid Stability Predictions (RaSP),
using their open-source code [1].

From a bird's eye view, RaSP is a black box that starts
with a collection of wildtype pdb files, and assesses the
stability of a (single-site) mutant. For the informed reader,
this black box can be considered a drop-in replacement of FoldX,
or Rosetta.

[1] “Rapid Protein Stability Prediction Using Deep Learning Representations.”
Blaabjerg, Lasse M, Maher M Kassem, Lydia L Good, Nicolas Jonsson,
Matteo Cagiada, Kristoffer E Johansson, Wouter Boomsma, Amelie Stein,
and Kresten Lindorff-Larsen.  Edited by José D Faraldo-Gómez,
Detlef Weigel, Nir Ben-Tal, and Julian Echave. eLife 12
(May 2023): e82593. https://doi.org/10.7554/eLife.82593.

"""

from __future__ import annotations

from pathlib import Path
from time import time
from typing import List, Union
from uuid import uuid4

import numpy as np
import torch

from poli.core.abstract_isolated_function import AbstractIsolatedFunction
from poli.core.util.proteins.mutations import find_closest_wildtype_pdb_file_to_mutant
from poli.core.util.proteins.pdb_parsing import (
    parse_pdb_as_residue_strings,
    parse_pdb_as_residues,
)
from poli.core.util.proteins.rasp import (
    RaspInterface,
    load_cavity_and_downstream_models,
)

RASP_NUM_ENSEMBLE = 10
if torch.cuda.is_available():
    RASP_DEVICE = "cuda"
else:
    RASP_DEVICE = "cpu"

THIS_DIR = Path(__file__).parent.resolve()
HOME_DIR = THIS_DIR.home()
RASP_DIR = HOME_DIR / ".poli_objectives" / "rasp"
RASP_DIR.mkdir(parents=True, exist_ok=True)

# This is the folder where all the files
# generated by RaSP will be stored.
# Feel free to change it if you want
# to keep the files somewhere else by
# passing tmp_folder to the black box.
# TODO: what happens if the user is on Windows?

# As a brief summary, this is what RaSP will
# do inside this folder: it will create 4
# subfolders inside it, one for each step
# of the pipeline (raw, cleaned, parsed, output).
#
#  - raw contains a version of the PDBs with only
#    the chain of interest.
#
#  - cleaned contains a version of the PDBs with
#    only the atoms of interest.
#
#  - parsed contains an .npz file with the variables
#    of interest, and
#
#  - output contains the predictions of the model.
DEFAULT_TMP_PATH = Path("/tmp").resolve()


class RaspIsolatedLogic(AbstractIsolatedFunction):
    """
    RaSP internal implementation.

    Parameters
    ----------
    wildtype_pdb_path : Union[Path, List[Path]]
        The path(s) to the wildtype PDB file(s), by default None.
    additive : bool, optional
        Whether we treat multiple mutations as additive, by default False.
        If you are interested in running this black box with multiple
        mutations, you should set this to True. Otherwise, it will
        raise an error if you pass a sequence with more than one mutation.
    chains_to_keep : List[str], optional
        The chains to keep in the PDB file(s), by default we
        keep the chain "A" for all pdbs passed.
    penalize_unfeasible_with: float, optional
        The value to return when the input is unfeasible, by default None, which means that we raise an error when
        an unfeasible sequence (e.g. one with a length different
        from the wildtypes) is passed.
    alphabet : List[str], optional
        The alphabet for the problem, by default we use
        the amino acid list provided in poli.core.util.proteins.defaults.
    experiment_id : str, optional
        The experiment ID, by default None.
    tmp_folder : Path, optional
        The temporary folder path, by default None.

    Methods
    -------
    _black_box(x, context=None)
        The main black box method that performs the computation, i.e.
        it computes the stability of the mutant(s) in x.
    _clean_wildtype_pdb_files()
        This function cleans the wildtype pdb files
        stored in self.wildtype_pdb_paths, using
        cached results if they exist.


    Raises
    ------
    AssertionError
        If wildtype_pdb_path is not provided.

    Notes
    -----
    - The wildtype_pdb_path can be a single Path object or a list of Path objects.
    - If chains_to_keep is not provided, it defaults to keeping chain A for all wildtypes.
    - If experiment_id is not provided, it is generated using the current timestamp and a random UUID.
    - If tmp_folder is not provided, it defaults to the default temporary path.
    """

    def __init__(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        additive: bool = False,
        chains_to_keep: List[str] = None,
        penalize_unfeasible_with: float | None = None,
        experiment_id: str = None,
        tmp_folder: Path = None,
        device: str | torch.device | None = None,
    ):
        """
        Initialize the RaSP Register object.

        Parameters:
        -----------
        wildtype_pdb_path : Union[Path, List[Path]]
            The path(s) to the wildtype PDB file(s).
        additive : bool, optional
            Whether we treat multiple mutations as additive, by default False.
            If you are interested in running this black box with multiple
            mutations, you should set this to True. Otherwise, it will
            raise an error if you pass a sequence with more than one mutation.
        chains_to_keep : List[str], optional
            The chains to keep in the PDB file(s), by default we
            keep the chain "A" for all pdbs passed.
        penalize_unfeasible_with: float, optional
            The value to return when the input is unfeasible, by default None, which means that we raise an error when
            an unfeasible sequence (e.g. one with a length different
            from the wildtypes) is passed.
        alphabet : List[str], optional
            The alphabet for the problem, by default we use
            the amino acid list provided in poli.core.util.proteins.defaults.
        experiment_id : str, optional
            The experiment ID, by default None.
        tmp_folder : Path, optional
            The temporary folder path, by default None.

        Raises:
        -------
        AssertionError
            If wildtype_pdb_path is not provided.

        Notes:
        ------
        - The wildtype_pdb_path can be a single Path object or a list of Path objects.
        - If chains_to_keep is not provided, it defaults to keeping chain A for all wildtypes.
        - If experiment_id is not provided, it is generated using the current timestamp and a random UUID.
        - If tmp_folder is not provided, it defaults to the default temporary path.
        """
        assert wildtype_pdb_path is not None, (
            "Missing required argument wildtype_pdb_file. "
            "Did you forget to pass it to create and into the black box?"
        )

        if isinstance(wildtype_pdb_path, str):
            wildtype_pdb_path = Path(wildtype_pdb_path.strip())

        if isinstance(wildtype_pdb_path, Path):
            wildtype_pdb_path = [wildtype_pdb_path]

        if isinstance(wildtype_pdb_path, list):
            if isinstance(wildtype_pdb_path[0], str):
                # Assuming that wildtype_pdb_path is a list of strings
                wildtype_pdb_path = [Path(x.strip()) for x in wildtype_pdb_path]
            elif isinstance(wildtype_pdb_path[0], Path):
                pass

        # By this point, we can make sure the wildtype_pdb_path
        # is a list of Path objects.
        assert all([x.exists() for x in wildtype_pdb_path]), (
            "One of the wildtype PDBs does not exist. "
            "Please check the path and try again."
        )

        self.wildtype_pdb_paths = wildtype_pdb_path

        self.wildtype_resiudes = [
            parse_pdb_as_residues(pdb_file) for pdb_file in wildtype_pdb_path
        ]

        self.wildtype_amino_acids = [
            parse_pdb_as_residue_strings(pdb_file) for pdb_file in wildtype_pdb_path
        ]

        self.wildtype_residue_strings = [
            "".join(amino_acids) for amino_acids in self.wildtype_amino_acids
        ]

        # Validating the chains to keep
        if isinstance(chains_to_keep, type(None)):
            # Defaulting to always keeping chain A.
            chains_to_keep = ["A"] * len(self.wildtype_pdb_paths)

        if isinstance(chains_to_keep, str):
            chains_to_keep = [chains_to_keep] * len(self.wildtype_pdb_paths)

        if isinstance(chains_to_keep, list):
            assert len(chains_to_keep) == len(self.wildtype_pdb_paths), (
                "The number of chains to keep must be the same as the number of wildtypes."
                " You can specify a single chain to keep for all wildtypes, or a list of chains."
            )

        # At this point, we are sure that chains_to_keep is a list of strings
        self.chains_to_keep = chains_to_keep

        self.penalize_unfeasible_with = penalize_unfeasible_with

        if experiment_id is None:
            experiment_id = f"{int(time())}_{str(uuid4())[:8]}"
        self.experiment_id = experiment_id

        self.tmp_folder = tmp_folder if tmp_folder is not None else DEFAULT_TMP_PATH

        # We need to preprocess these pdbs if they haven't
        # been preprocessed. This calls for a new folder
        # in tmp where _all_ pdbs are preprocessed and
        # stored... (?)
        self.clean_wildtype_pdb_files = None

        # After this, self.clean_wildtype_pdb_files is a
        # list of Path objects.
        self.working_dir = self.tmp_folder / "RaSP_tmp_files" / self.experiment_id
        self.rasp_interface = RaspInterface(working_dir=self.working_dir)
        self._clean_wildtype_pdb_files()

        x0_pre_array = []
        for clean_wildtype_pdb_file in self.clean_wildtype_pdb_files:
            # Loads up the wildtype pdb files as strings
            wildtype_string = self.parse_pdb_as_residue_strings(clean_wildtype_pdb_file)
            x0_pre_array.append(list(wildtype_string))

        # Padding all of them to the longest sequence
        max_len = max([len(x) for x in x0_pre_array])
        x0_pre_array = [x + [""] * (max_len - len(x)) for x in x0_pre_array]

        self.x0 = np.array(x0_pre_array)
        self.additive = additive
        self.device = device if device is not None else RASP_DEVICE

    def _clean_wildtype_pdb_files(self):
        """
        This function cleans the wildtype pdb files
        stored in self.wildtype_pdb_paths, using
        cached results if they exist.
        """

        # We make sure the PDBs are cleaned.
        # TODO: add chain to the name of the file
        for wildtype_pdb_path, chain_to_keep in zip(
            self.wildtype_pdb_paths, self.chains_to_keep
        ):
            if not (
                self.working_dir
                / "raw"
                / f"{wildtype_pdb_path.stem}_{chain_to_keep}.pdb"
            ).exists():
                self.rasp_interface.raw_pdb_to_unique_chain(
                    wildtype_pdb_path, chain=chain_to_keep
                )

            if not (
                self.working_dir
                / "cleaned"
                / f"{wildtype_pdb_path.stem}_{chain_to_keep}_clean.pdb"
            ).exists():
                self.rasp_interface.unique_chain_to_clean_pdb(wildtype_pdb_path)

            if not (
                self.working_dir
                / "parsed"
                / f"{wildtype_pdb_path.stem}_{chain_to_keep}_clean_coordinate_features.npz"
            ).exists():
                self.rasp_interface.cleaned_to_parsed_pdb(wildtype_pdb_path)

        self.clean_wildtype_pdb_files = [
            self.working_dir
            / "cleaned"
            / f"{wildtype_pdb_path.stem}_query_protein_uniquechain_clean.pdb"
            for wildtype_pdb_path in self.wildtype_pdb_paths
        ]

    def parse_pdb_as_residue_strings(self, pdb_file: Path) -> List[str]:
        return parse_pdb_as_residue_strings(pdb_file)

    def _compute_mutant_residue_string_ddg(
        self, mutant_residue_string: str
    ) -> np.ndarray:
        for i, char in enumerate(mutant_residue_string):
            if char not in self.rasp_interface.alphabet:
                raise ValueError(
                    f"Invalid residue {char} at position {i} in the mutant "
                    "residue string. Please make sure that all residues are "
                    f"in the alphabet: {self.rasp_interface.alphabet}."
                )
        try:
            (
                closest_wildtype_pdb_file,
                hamming_distance,
            ) = find_closest_wildtype_pdb_file_to_mutant(
                self.clean_wildtype_pdb_files,
                mutant_residue_string,
                return_hamming_distance=True,
            )
        except ValueError as e:
            # This means that the mutant is unfeasible
            if self.penalize_unfeasible_with is not None:
                # We return the penalization value with a minus sign
                # since we are maximizing the stability (i.e. there
                # is a sign flip at the end of the __call__ method).
                return np.array([-self.penalize_unfeasible_with])
            else:
                raise ValueError(
                    "The mutant's length does not match any of the wildtypes. If "
                    "you want to proceed by penalizing this behavior, you should "
                    "set penalize_unfeasible_with to a value different from None "
                    "in the create method/black box constructor of RaSP."
                ) from e

        if hamming_distance > 1 and not self.additive:
            raise ValueError(
                "RaSP is only able to simulate single mutations."
                " If you want to simulate multiple mutations,"
                " you should set additive=True in the create method"
                " or in the black box of RaSP."
            )

        # Loading the models in preparation for inference
        cavity_model_net, ds_model_net = load_cavity_and_downstream_models(
            device=self.device
        )
        dataset_key = "predictions"

        df_structure = self.rasp_interface.create_df_structure(
            closest_wildtype_pdb_file,
            mutant_residue_strings=[mutant_residue_string],
        )

        # STEP 3:
        # Predicting
        df_ml = self.rasp_interface.predict(
            cavity_model_net,
            ds_model_net,
            df_structure,
            dataset_key,
            RASP_NUM_ENSEMBLE,
            RASP_DEVICE,
        )

        sliced_values_for_mutant = df_ml["score_ml"][
            df_ml["mutant_residue_string"] == mutant_residue_string
        ].values
        if self.additive:
            if hamming_distance > 0:
                assert sliced_values_for_mutant.shape[0] == hamming_distance, (
                    " The number of predictions made for this mutant"
                    " is not equal to the Hamming distance between the"
                    " mutant and the wildtype.\n"
                    "This is an internal error in `poli`. Please report"
                    " this issue by referencing the RaSP problem.\n"
                    " https://github.com/MachineLearningLifeScience/poli/issues"
                )

                result = np.sum(sliced_values_for_mutant, keepdims=True)
            else:
                result = sliced_values_for_mutant
        else:
            result = sliced_values_for_mutant

        return result

    def __call__(self, x, context=None):
        """
        Computes the stability of the mutant(s) in x.

        Parameters
        ----------
        x : np.ndarray
            Input array of shape [b, L] containing strings.
        context : dict, optional
            Additional context information (default is None).

        Returns
        -------
        y : np.ndarray
            The stability of the mutant(s) in x.

        Notes
        -----
        - x is a np.array[str] of shape [b, L], where L is the length
          of the longest sequence in the batch, and b is the batch size.
          We process it by concantenating the array into a single string,
          where we assume the padding to be an empty string (if there was any).
          Each of these x_i's will be matched to the wildtype in
          self.wildtype_residue_strings with the lowest Hamming distance.
        """
        # We need to find the closest wildtype to each of the
        # sequences in x. For this, we need to compute the
        # Hamming distance between each of the sequences in x
        # and each of the wildtypes in self.wildtype_residue_strings.

        # closest_wildtypes will be a dictionary
        # of the form {wildtype_path: List[str] of mutations}
        # closest_wildtypes = defaultdict(list)
        # mutant_residue_strings = []
        # mutant_residue_to_hamming_distances = dict()
        results = []
        for x_i in x:
            # Assuming x_i is an array of strings
            mutant_residue_string = "".join(x_i)
            result = self._compute_mutant_residue_string_ddg(mutant_residue_string)
            results.append(result)

        return -np.array(results).reshape(-1, 1)


if __name__ == "__main__":
    from poli.core.registry import register_isolated_function

    register_isolated_function(
        RaspIsolatedLogic,
        name="rasp__isolated",
        conda_environment_name="poli__rasp",
        force=True,
    )



"""Registers the toy continuous problem objective function and factory.

This is a registration script for the toy continuous
objectives which are usually used to benchmark continuous
optimization algorithms in several dimensions.

We focus on the ones that allow us to specify the
dimensionality of the problem [1].

The problem is registered as 'toy_continuous_problem',
and it uses a conda environment called 'poli__base'
(see the environment.yml file in this folder).
"""

from typing import List

import numpy as np

from poli.core.abstract_black_box import AbstractBlackBox
from poli.core.abstract_problem_factory import AbstractProblemFactory
from poli.core.black_box_information import BlackBoxInformation
from poli.core.problem import Problem
from poli.core.util.seeding import seed_python_numpy_and_torch

from .toy_continuous_problem import POSSIBLE_FUNCTIONS, ToyContinuousProblem


class ToyContinuousBlackBox(AbstractBlackBox):
    """
    A black box implementation for evaluating the Toy Continuous Problem.

    Parameters
    ----------
    function_name : str
        The name of the toy continuous function to evaluate, by default None.
    n_dimensions : int
        The number of dimensions for the toy continuous function, by default 2.
    embed_in : int, optional
        If not None, the continuous problem is randomly embedded in this dimension.
        By default, None.
    dimensions_to_embed_in: List[int], optional
        The dimensions in which to embed the problem, by default None. Only has an effect if embed_in is not None.
    batch_size : int, optional
        The batch size for parallel evaluation, by default None.
    parallelize : bool, optional
        Whether to parallelize the evaluation, by default False.
    num_workers : int, optional
        The number of workers for parallel evaluation, by default None.
    evaluation_budget : int, optional
        The maximum number of evaluations, by default float("inf").

    Attributes
    ----------
    function_name : str
        The name of the toy continuous function.
    n_dimensions : int
        The number of dimensions for the toy continuous function.
    embed_in : int
        The dimension in which to embed the problem.
    function : ToyContinuousProblem
        The toy continuous problem instance.
    bounds : Tuple[np.ndarray, np.ndarray]
        The lower and upper bounds for the toy continuous problem.

    Methods
    -------
    _black_box(x, context=None)
        Evaluates the toy continuous problem on a continuous input x.

    """

    def __init__(
        self,
        function_name: str,
        n_dimensions: int = 2,
        embed_in: int = None,
        dimensions_to_embed_in: List[int] = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
    ):

        assert (
            function_name in POSSIBLE_FUNCTIONS
        ), f"'{function_name}' is not a valid function name. Expected it to be one of {POSSIBLE_FUNCTIONS}."

        self.function_name = function_name
        self.n_dimensions = n_dimensions
        self.embed_in = embed_in
        self.function = ToyContinuousProblem(
            function_name,
            n_dims=n_dimensions,
            embed_in=embed_in,
            dimensions_to_embed_in=dimensions_to_embed_in,
        )
        self.bounds = self.function.limits
        self.optima = self.function.optima

        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )

    def _black_box(self, x: np.ndarray, context: dict = None) -> np.ndarray:
        """
        Evaluates the toy continuous problem on a continuous input x.

        Parameters
        ----------
        x : np.ndarray
            The input to evaluate (expected to be an array [b, L] of floats).
        context : dict, optional
            The context of the evaluation, by default None.

        Returns
        -------
        y : np.ndarray
            The evaluation of the toy continuous problem.

        Raises
        ------
        ValueError
            If the input x is not of type float.

        """
        if not x.dtype.kind == "f":
            raise ValueError("Expected a batch of floats. ")

        return self.function(x)

    def get_black_box_info(self) -> BlackBoxInformation:
        return BlackBoxInformation(
            name="toy_continuous_problem",
            max_sequence_length=(
                self.embed_in if self.embed_in is not None else self.n_dimensions
            ),
            aligned=True,
            fixed_length=True,
            deterministic=True,
            alphabet=None,
            log_transform_recommended=False,
            discrete=False,
            padding_token=None,
        )


class ToyContinuousProblemFactory(AbstractProblemFactory):
    def create(
        self,
        function_name: str,
        n_dimensions: int = 2,
        embed_in: int = None,
        dimensions_to_embed_in: List[int] = None,
        seed: int = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        force_isolation: bool = False,
    ) -> Problem:
        """
        Creates a new instance of the toy continuous problem.

        Parameters
        ----------
        function_name : str
            The name of the toy continuous function to evaluate.
        n_dimensions : int, optional
            The number of dimensions for the toy continuous function, by default 2.
        embed_in : int, optional
            If not None, the continuous problem is randomly embedded in this dimension.
            By default, None.
        dimensions_to_embed_in: List[int], optional
            The dimensions in which to embed the problem, by default None. Only has an effect if embed_in is not None.
        seed : int, optional
            The seed for the random number generator, by default None.
        batch_size : int, optional
            The batch size for simultaneous execution, by default None.
        parallelize : bool, optional
            Flag indicating whether to parallelize execution, by default False.
        num_workers : int, optional
            The number of workers for parallel execution, by default None.
        evaluation_budget:  int, optional
            The maximum number of function evaluations. Default is infinity.

        Returns
        -------
        problem : Problem
            The problem instance containing the black box and an initial point.

        Raises
        ------
        ValueError
            If the function_name is not one of the valid functions.
        """
        assert (
            function_name in POSSIBLE_FUNCTIONS
        ), f"'{function_name}' is not a valid function name. Expected it to be one of {POSSIBLE_FUNCTIONS}."

        if seed is not None:
            seed_python_numpy_and_torch(seed)

        f = ToyContinuousBlackBox(
            function_name=function_name,
            n_dimensions=n_dimensions,
            embed_in=embed_in,
            dimensions_to_embed_in=dimensions_to_embed_in,
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
        x0 = f.function.x0

        return Problem(f, x0)



from poli.core.chemistry.data_packages import RandomMoleculesDataPackage
from poli.core.chemistry.tdc_black_box import TDCBlackBox
from poli.core.problem import Problem


class TDCProblem(Problem):
    def __init__(
        self, black_box: TDCBlackBox, x0, data_package=None, strict_validation=True
    ):
        if data_package is None:
            data_package = RandomMoleculesDataPackage(black_box.string_representation)

        super().__init__(
            black_box=black_box,
            x0=x0,
            data_package=data_package,
            strict_validation=strict_validation,
        )



from pathlib import Path
from typing import List, Union

import numpy as np

from poli.core.proteins.foldx_isolated_function import FoldxIsolatedFunction
from poli.core.registry import register_isolated_function
from poli.core.util.proteins.foldx import FoldxInterface
from poli.core.util.proteins.mutations import find_closest_wildtype_pdb_file_to_mutant


class FoldXStabilitityAndSASAIsolatedLogic(FoldxIsolatedFunction):
    """
    A black box implementation for computing stability and solvent accessible surface area (SASA) score using FoldX.

    Parameters
    -----------
    wildtype_pdb_path : Union[Path, List[Path]]
        The path(s) to the wildtype PDB file(s). Default is None.
    alphabet : List[str], optional
        The alphabet of amino acids. Default is None.
    experiment_id : str, optional
        The ID of the experiment. Default is None.
    tmp_folder : Path, optional
        The path to the temporary folder. Default is None.
    eager_repair : bool, optional
        Whether to perform eager repair. Default is False.
    verbose : bool, optional
        Flag indicating whether we print the output from FoldX. Default is False.

    Notes
    -----
    We expect the user to have FoldX v5.0 installed and compiled.
    More specifically, we expect a binary file called foldx to be
    in the path ~/foldx/foldx.
    """

    def __init__(
        self,
        wildtype_pdb_path: Union[Path, List[Path]],
        experiment_id: str = None,
        tmp_folder: Path = None,
        eager_repair: bool = False,
        verbose: bool = False,
    ):
        super().__init__(
            wildtype_pdb_path=wildtype_pdb_path,
            experiment_id=experiment_id,
            tmp_folder=tmp_folder,
            eager_repair=eager_repair,
            verbose=verbose,
        )

    def __call__(self, x: np.ndarray, context: None) -> np.ndarray:
        """Computes the SASA score for a given mutation x.

        Runs the given input x and pdb files provided
        in the context through FoldX and returns the
        total SASA score.

        Parameters
        -----------
        x : np.ndarray
            The input array representing the mutations.
        context : None
            The context for the black box computation.

        Returns
        --------
        np.ndarray
            The computed SASA score(s) as a numpy array.
        """
        # TODO: add support for multiple mutations.
        # For now, we assume that the batch size is
        # always 1.
        assert x.shape[0] == 1, "We only support single mutations for now. "

        # Create a working directory for this function call
        working_dir = self.create_working_directory()

        # Given x, we simply define the
        # mutations to be made as a mutation_list.txt
        # file.
        mutations_as_strings = [
            "".join([amino_acid for amino_acid in x_i]) for x_i in x
        ]

        # We find the associated wildtype to this given
        # mutation. This is done by minimizing the
        # Hamming distance between the mutated residue
        # string and the wildtype residue strings of
        # all the PDBs we have.
        # TODO: this assumes a batch size of 1.
        wildtype_pdb_file = find_closest_wildtype_pdb_file_to_mutant(
            self.wildtype_pdb_paths, mutations_as_strings[0]
        )

        foldx_interface = FoldxInterface(working_dir, self.verbose)
        stability, sasa_score = foldx_interface.compute_stability_and_sasa(
            pdb_file=wildtype_pdb_file,
            mutations=mutations_as_strings,
        )

        return np.array([stability, sasa_score]).reshape(-1, 2)


if __name__ == "__main__":
    register_isolated_function(
        FoldXStabilitityAndSASAIsolatedLogic,
        name="foldx_stability_and_sasa__isolated",
        conda_environment_name="poli__protein",
    )



from poli.core.abstract_black_box import AbstractBlackBox


class IsolatedBlackBox(AbstractBlackBox):
    def __init__(
        self,
        name: str = None,
        batch_size: int = None,
        parallelize: bool = False,
        num_workers: int = None,
        evaluation_budget: int = float("inf"),
        **kwargs_for_black_box,
    ):

        super().__init__(
            batch_size=batch_size,
            parallelize=parallelize,
            num_workers=num_workers,
            evaluation_budget=evaluation_budget,
        )
